# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2019-, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from abc import abstractmethod
from typing import Dict, List, Optional, Set, Tuple

import torch as t
import torch.nn as nn

from nemo.core import DeviceType, ModuleType, NeuralModule, OperationMode, WeightShareTransform
from nemo.utils.helpers import get_cuda_device, rgetattr, rsetattr


class TrainableNM(NeuralModule, nn.Module):
    """A helper Base class for NeuralModule's based on Pytorch's nn.Module.

    If you have a Pytorch class which derives from nn.Module you can
    covert it into a NeuralModule, by replacing inheriting from this class
    instead

    Your constructor then should look like this:

    .. code-block:: python

      def __init__(self):
        super().__init__()
        .... # your code

    Then make sure that your forward(..) method accepts arguments named like
    input ports.

    Args:
        pretrained_model_name (str): name of pretrained model to use in order
            to initialize this neural module

    """

    def __init__(self, pretrained_model_name=None, name=None):
        # Initialize nn.Module first - important for the inspect during the init_params collection.
        nn.Module.__init__(self)  # For PyTorch API
        NeuralModule.__init__(self, name)  # For NeuralModule API

        # Unfrozen by default.
        self._frozen = False

        # Set module type.
        self._type = ModuleType.trainable

        self._device = get_cuda_device(self.placement)

        # Store pretrained model name (to be removed/changed)
        self._pretrained_model_name = pretrained_model_name

    def __call__(self, *input, force_pt=False, **kwargs):
        pt_call = len(input) > 0 or force_pt
        if pt_call:
            return nn.Module.__call__(self, *input, **kwargs)
        else:
            return NeuralModule.__call__(self, **kwargs)

    @t.jit.ignore
    def get_weights(self):
        result = dict()
        for name, parameter in self.named_parameters():
            result[name] = (parameter, parameter.requires_grad)
        return result

    @t.jit.ignore
    def set_weights(self, name2weight, name2name_and_transform=None):
        if name2weight is not None and len(name2weight) > 0:
            if name2name_and_transform is None:
                self.load_state_dict({key: name2weight[key][0] for key in name2weight.keys()})
            else:
                raise NotImplementedError("Transforms are not currently supported for set_weights")

    @t.jit.ignore
    def tie_weights_with(self, module, weight_names, name2name_and_transform=None):
        if module is None:
            raise ValueError("Module to tie weights can't be None")
        if weight_names is None or len(weight_names) == 0:
            raise ValueError("Please provide weight names to tie")

        if name2name_and_transform is None:
            for name in weight_names:
                rsetattr(self, name, nn.Parameter(rgetattr(module, name)))
        else:
            for self_w_name in weight_names:
                if self_w_name in name2name_and_transform:
                    if name2name_and_transform[self_w_name][1] == WeightShareTransform.SAME:
                        rsetattr(
                            self, self_w_name, nn.Parameter(rgetattr(module, name2name_and_transform[self_w_name][0])),
                        )
                    elif name2name_and_transform[self_w_name][1] == WeightShareTransform.TRANSPOSE:
                        raise NotImplementedError("Sorry, currently this is not implemented.")
                else:
                    rsetattr(self, self_w_name, nn.Parameter(rgetattr(module, self_w_name)))

    @t.jit.ignore
    def save_to(self, path):
        # t.save(self._pt_module.state_dict(), path)
        t.save(self.state_dict(), path)

    @t.jit.ignore
    def restore_from(self, path, local_rank=0):
        # self._pt_module.load_state_dict(t.load(path))
        if self.placement == DeviceType.AllGpu:
            load_device = f"cuda:{local_rank}"
        else:
            load_device = self._device
        self.load_state_dict(t.load(path, map_location=load_device))

    @t.jit.ignore
    def freeze(self, weights=None):
        if hasattr(self, "_pt_module"):
            for name, param in self._pt_module.named_parameters():
                if weights is None or name in weights:
                    param.requires_grad = False
        else:
            for name, param in self.named_parameters():
                if weights is None or name in weights:
                    param.requires_grad = False
        # Freeze.
        self._frozen = True

    @t.jit.ignore
    def unfreeze(self, weights=None):
        if hasattr(self, "_pt_module"):
            for name, param in self._pt_module.named_parameters():
                if weights is None or name in weights:
                    param.requires_grad = True
        else:
            for name, param in self.named_parameters():
                if weights is None or name in weights:
                    param.requires_grad = True
        # Unfreeze.
        self._frozen = False

    @t.jit.ignore
    def is_frozen(self) -> bool:
        """ Returns:
                True/False depending whether there are any frozen weights or not.
        """
        return self._frozen

    @property
    def operation_mode(self):
        """ Returns the operation mode. """
        return self._operation_mode

    @operation_mode.setter
    def operation_mode(self, operation_mode: OperationMode):
        """ Sets the operation mode. Additionally, depending on the mode, runs PyTorch train()/eval()"""
        # Copy operation mode.
        self._operation_mode = operation_mode

        # Set PyTorch train/eval flag.
        if operation_mode == OperationMode.evaluation:
            self.eval()
        else:
            # & OperationMode.training & OperationMode.both.
            # In the latter not sure what the user will do, so collect the gradients "just in case".
            self.train()

    @property
    def num_weights(self):
        return sum(p.numel() for p in self.parameters() if p.requires_grad)


class NonTrainableNM(NeuralModule):
    def __init__(self, name=None):
        NeuralModule.__init__(self, name)  # For NeuralModule API
        self._device = get_cuda_device(self.placement)
        # Set module type.
        self._type = ModuleType.nontrainable

    def __call__(self, force_pt=False, *input, **kwargs):
        pt_call = len(input) > 0 or force_pt
        if pt_call:
            return self.forward(*input, **kwargs)
        else:
            return NeuralModule.__call__(self, **kwargs)

    def forward(self, *input):
        """Defines the computation performed at every call.

        Should be overridden by all subclasses.
        """
        raise NotImplementedError

    def get_weights(self) -> Optional[Dict[(str, bool)]]:
        None

    def set_weights(
        self,
        name2weight: Dict[(str, Tuple[str, bool])],
        name2name_and_transform: Dict[(str, Tuple[str, WeightShareTransform])] = None,
    ):
        pass

    def tie_weights_with(
        self,
        module,
        weight_names=List[str],
        name2name_and_transform: Dict[(str, Tuple[str, WeightShareTransform])] = None,
    ):
        pass

    def save_to(self, path: str):
        pass

    def restore_from(self, path: str, local_rank: int = 0):
        pass

    def freeze(self, weights: Set[str] = None):
        pass

    def unfreeze(self, weights: Set[str] = None):
        pass

    @property
    def num_weights(self):
        return 0


class DataLayerNM(NeuralModule):
    """A helper Base class for creating Pytorch-based data layers.
    You must implement __len__ method to return dataset size and
    data_iterator property to return iterator over the dataset.
    """

    def __init__(self, name=None):
        NeuralModule.__init__(self, name)  # For NeuralModule API

        # Set module type.
        self._type = ModuleType.datalayer

        self._device = get_cuda_device(self.placement)

        # if 'batch_size' not in kwargs:
        #    logging.warning("No batch_size specified in the data layer. "
        #                    "Setting batch_size to 1.")
        #    kwargs['batch_size'] = 1

        # Set default values of variables used by trained/passed to DataLoader.
        # NOTE: That also means that those are parameters of DataLoader/trainer, not DataLayer.
        # Thus those fields will be removed from DataLayer and moved to trainer configuration
        # (when the time for that will come;))
        self._batch_size = 1
        self._num_workers = os.cpu_count()  # Use all CPUs by default.
        self._shuffle = False  # Don't shuffle by default.
        self._pin_memory = False

    @property
    def input_ports(self):
        """DataLayer by definition does not have any input ports.

            Returns:
                An empty dictionary.
        """
        return {}

    def get_weights(self):
        # logging.warning(
        #     "Data Layer does not have any weights to return. "
        #     "This get_weights call returns None."
        # )
        return None

    def set_weights(self, name2weight: Dict[(str, bool)], name2name_and_transform):
        # logging.warning(
        #     "Data Layer does not have any weights to set. "
        #     "This set_weights call is ignored."
        # )
        return None

    def tie_weights_with(self, module, weight_names):
        # logging.warning(
        #     "Data Layer does not have any weights to tie. "
        #     "This tie_weights_with call is ignored."
        # )
        return None

    def save_to(self, path):
        # logging.warning(
        #     "Data Layer does not have any state to save. "
        #     "This save_to call is ignored."
        # )
        return None

    def restore_from(self, path):
        raise NotImplementedError("Data Layer could not be restored from any saved state.")
        return None

    def freeze(self, weights: Set[str] = None):
        # logging.warning(
        #     "Data Layer does not have any weights to freeze. "
        #     "This freeze call is ignored."
        # )
        return None

    def unfreeze(self, weights: Set[str] = None):
        # logging.warning(
        #     "Data Layer does not have any weights to unfreeze. "
        #     "This unfreeze call is ignored."
        # )
        return None

    @property
    def num_weights(self):
        return 0

    @abstractmethod
    def __len__(self):
        """Dataset size"""
        pass

    @property
    @abstractmethod
    def dataset(self):
        """Should return an instance of torch.utils.data.Dataset. Should
        implement
        either this or `data_iterator`. If this is implemented, `data_iterator`
        should return None."""
        pass

    @property
    def data_iterator(self):
        """"Iterator over the dataset. It is a good idea to return
        torch.utils.data.DataLoader here. Should implement either this or
        `dataset`.
        If this is implemented, `dataset` property should return None.
        """
        return None

    @property
    def batch_size(self):
        """ Property returning the batch size. """
        return self._batch_size

    # @batch_size.setter
    # def batch_size(self, bs):
    #    """ Property setting the batch size. """
    #    self._batch_size = bs

    @property
    def shuffle(self):
        """ Property returning the shuffle flag. """
        return self._shuffle

    # @shuffle.setter
    # def shuffle(self, sh):
    #    """ Property setting the shuffle flag. """
    #     self._shuffle = sh

    @property
    def num_workers(self):
        """ Property returning the number of workers. """
        return self._num_workers

    # @num_workers.setter
    # def num_workers(self, nw):
    #    """ Property setting the number of workers. """
    #    self._num_workers = nw

    @property
    def pin_memory(self):
        """ Property returning the pin memory flag. """
        return self._pin_memory


class LossNM(NeuralModule):
    """A helper Base class for creating Pytorch-based loss function modules.
    You must implement _loss_function method.
    """

    def __init__(self, name=None):
        NeuralModule.__init__(self, name)  # For NeuralModule API

        # Set module type.
        self._type = ModuleType.loss

        self._device = get_cuda_device(self.placement)

    def get_weights(self):
        # logging.warning(
        #     "Loss function module does not have any weights "
        #      "to return. This get_weights call returns None.")
        return None

    def set_weights(self, name2weight: Dict[(str, bool)], name2name_and_transform):
        # logging.warning(
        #     "Loss function module does not have any weights to set. "
        #     "This set_weights call is ignored."
        # )
        return None

    def tie_weights_with(self, module, weight_names):
        # logging.warning(
        #     "Loss function module does not have any weights to tie. "
        #     "This tie_weights_with call is ignored."
        # )
        return None

    def save_to(self, path):
        # logging.warning(
        #     "Loss function module does not have any state to save. "
        #     "This save_to call is ignored."
        # )
        return None

    def restore_from(self, path):
        raise NotImplementedError("Loss function module could not be restored from any saved state.")
        return None

    def freeze(self, weights: Set[str] = None):
        # logging.warning(
        #     "Loss function module does not have any weights to freeze. "
        #     "This freeze call is ignored."
        # )
        return None

    def unfreeze(self, weights: Set[str] = None):
        # logging.warning(
        #     "Loss function module does not have any weights to "
        #     "unfreeze. This unfreeze call is ignored."
        # )
        return None

    @property
    def num_weights(self):
        return 0

    @abstractmethod
    def _loss_function(self, **kwargs):
        pass

    def __call__(self, force_pt=False, *input, **kwargs):
        if force_pt:
            return self._loss_function(**kwargs)
        else:
            return NeuralModule.__call__(self, **kwargs)



# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import os

import numpy as np
import torch
from torch import nn


class FastSpeechDataset:
    def __init__(self, audio_dataset, durs_dir):
        self._audio_dataset = audio_dataset
        self._durs_dir = durs_dir

    def __getitem__(self, index):
        audio, audio_len, text, text_len = self._audio_dataset[index]
        dur_true = torch.tensor(np.load(os.path.join(self._durs_dir, f'{index}.npy'))).long()
        return dict(audio=audio, audio_len=audio_len, text=text, text_len=text_len, dur_true=dur_true)

    def __len__(self):
        return len(self._audio_dataset)


class LengthRegulator(nn.Module):
    """Length Regulator."""

    def __init__(self, encoder_output_size, duration_predictor_filter_size, duration_predictor_kernel_size, dropout):
        super(LengthRegulator, self).__init__()

        self.duration_predictor = DurationPredictor(
            input_size=encoder_output_size,
            filter_size=duration_predictor_filter_size,
            kernel=duration_predictor_kernel_size,
            conv_output_size=duration_predictor_filter_size,
            dropout=dropout,
        )

    def forward(self, encoder_output, encoder_output_mask, target=None, alpha=1.0, mel_max_length=None):
        duration_predictor_output = self.duration_predictor(encoder_output, encoder_output_mask)

        if self.training:
            output, dec_pos = self.get_output(encoder_output, target, alpha, mel_max_length)
        else:
            duration_predictor_output = torch.clamp_min(torch.exp(duration_predictor_output) - 1, 0)

            output, dec_pos = self.get_output(encoder_output, duration_predictor_output, alpha)

        return output, dec_pos, duration_predictor_output

    @staticmethod
    def get_output(encoder_output, duration_predictor_output, alpha, mel_max_length=None):
        output = list()
        dec_pos = list()

        for i in range(encoder_output.size(0)):
            repeats = duration_predictor_output[i].float() * alpha
            repeats = torch.round(repeats).long()
            output.append(torch.repeat_interleave(encoder_output[i], repeats, dim=0))
            dec_pos.append(torch.from_numpy(np.indices((output[i].shape[0],))[0] + 1))

        output = torch.nn.utils.rnn.pad_sequence(output, batch_first=True)
        dec_pos = torch.nn.utils.rnn.pad_sequence(dec_pos, batch_first=True)

        dec_pos = dec_pos.to(output.device, non_blocking=True)

        if mel_max_length:
            output = output[:, :mel_max_length]
            dec_pos = dec_pos[:, :mel_max_length]

        return output, dec_pos


class ConvTranspose(nn.Module):
    """Convolution Module with transposes of last two dimensions."""

    def __init__(
        self, in_channels, out_channels, kernel_size=1, stride=1, padding=0, dilation=1, bias=True, w_init='relu'
    ):
        super(ConvTranspose, self).__init__()

        self.conv = nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            bias=bias,
        )

        nn.init.xavier_uniform_(self.conv.weight, gain=nn.init.calculate_gain(w_init))

    def forward(self, x):
        x = x.contiguous().transpose(1, 2)
        x = self.conv(x)
        x = x.contiguous().transpose(1, 2)

        return x


class DurationPredictor(nn.Module):
    """Duration Predictor."""

    def __init__(self, input_size, filter_size, kernel, conv_output_size, dropout):
        super(DurationPredictor, self).__init__()

        self.input_size = input_size
        self.filter_size = filter_size
        self.kernel = kernel
        self.conv_output_size = conv_output_size
        self.dropout = dropout

        self.conv_layer = nn.Sequential(
            collections.OrderedDict(
                [
                    (
                        "conv1d_1",
                        ConvTranspose(self.input_size, self.filter_size, kernel_size=self.kernel, padding=1),
                    ),
                    ("relu_1", nn.ReLU()),
                    ("layer_norm_1", nn.LayerNorm(self.filter_size)),
                    ("dropout_1", nn.Dropout(self.dropout)),
                    (
                        "conv1d_2",
                        ConvTranspose(self.filter_size, self.filter_size, kernel_size=self.kernel, padding=1),
                    ),
                    ("relu_2", nn.ReLU()),
                    ("layer_norm_2", nn.LayerNorm(self.filter_size)),
                    ("dropout_2", nn.Dropout(self.dropout)),
                ]
            )
        )

        self.linear_layer = nn.Linear(self.conv_output_size, 1, bias=True)

    def forward(self, encoder_output, encoder_output_mask):
        encoder_output = encoder_output * encoder_output_mask

        out = self.conv_layer(encoder_output)
        out = self.linear_layer(out)
        out = out * encoder_output_mask
        out = out.squeeze(-1)

        return out



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

"""
This file contains code artifacts adapted from the original implementation:
https://github.com/thu-coai/ConvLab-2/blob/master/convlab2/util/multiwoz/multiwoz_slot_trans.py
"""

__all__ = ['REF_USR_DA', 'REF_SYS_DA']

REF_USR_DA = {
    'Attraction': {
        'area': 'Area',
        'type': 'Type',
        'name': 'Name',
        'entrance fee': 'Fee',
        'address': 'Addr',
        'postcode': 'Post',
        'phone': 'Phone',
    },
    'Hospital': {'department': 'Department', 'address': 'Addr', 'postcode': 'Post', 'phone': 'Phone'},
    'Hotel': {
        'type': 'Type',
        'parking': 'Parking',
        'pricerange': 'Price',
        'internet': 'Internet',
        'area': 'Area',
        'stars': 'Stars',
        'name': 'Name',
        'stay': 'Stay',
        'day': 'Day',
        'people': 'People',
        'address': 'Addr',
        'postcode': 'Post',
        'phone': 'Phone',
    },
    'Police': {'address': 'Addr', 'postcode': 'Post', 'phone': 'Phone', 'name': 'Name'},
    'Restaurant': {
        'food': 'Food',
        'pricerange': 'Price',
        'area': 'Area',
        'name': 'Name',
        'time': 'Time',
        'day': 'Day',
        'people': 'People',
        'phone': 'Phone',
        'postcode': 'Post',
        'address': 'Addr',
    },
    'Taxi': {
        'leaveAt': 'Leave',
        'destination': 'Dest',
        'departure': 'Depart',
        'arriveBy': 'Arrive',
        'car type': 'Car',
        'phone': 'Phone',
    },
    'Train': {
        'destination': 'Dest',
        'day': 'Day',
        'arriveBy': 'Arrive',
        'departure': 'Depart',
        'leaveAt': 'Leave',
        'people': 'People',
        'duration': 'Time',
        'price': 'Ticket',
        'trainID': 'Id',
    },
}

REF_SYS_DA = {
    'Attraction': {
        'Addr': "address",
        'Area': "area",
        'Choice': "choice",
        'Fee': "entrance fee",
        'Name': "name",
        'Phone': "phone",
        'Post': "postcode",
        'Price': "pricerange",
        'Type': "type",
        'none': None,
        'Open': None,
    },
    'Hospital': {'Department': 'department', 'Addr': 'address', 'Post': 'postcode', 'Phone': 'phone', 'none': None},
    'Booking': {
        'Day': 'day',
        'Name': 'name',
        'People': 'people',
        'Ref': 'Ref',
        'Stay': 'stay',
        'Time': 'time',
        'none': None,
    },
    'Hotel': {
        'Addr': "address",
        'Area': "area",
        'Choice': "choice",
        'Internet': "internet",
        'Name': "name",
        'Parking': "parking",
        'Phone': "phone",
        'Post': "postcode",
        'Price': "pricerange",
        'Ref': "Ref",
        'Stars': "stars",
        'Type': "type",
        'none': None,
    },
    'Restaurant': {
        'Addr': "address",
        'Area': "area",
        'Choice': "choice",
        'Name': "name",
        'Food': "food",
        'Phone': "phone",
        'Post': "postcode",
        'Price': "pricerange",
        'Ref': "Ref",
        'none': None,
    },
    'Taxi': {
        'Arrive': "arriveBy",
        'Car': "taxi_types",
        'Depart': "departure",
        'Dest': "destination",
        'Leave': "leaveAt",
        'Phone': "taxi_phone",
        'none': None,
    },
    'Train': {
        'Arrive': "arriveBy",
        'Choice': "choice",
        'Day': "day",
        'Depart': "departure",
        'Dest': "destination",
        'Id': "trainID",
        'Leave': "leaveAt",
        'People': "people",
        'Ref': "Ref",
        'Time': "duration",
        'none': None,
        'Ticket': 'price',
    },
    'Police': {'Addr': "address", 'Post': "postcode", 'Phone': "phone"},
}



# Copyright (c) 2019 NVIDIA Corporation
import string
from typing import List, Optional

import frozendict

from nemo.collections.asr.parts import cleaners


class CharParser:
    """Functor for parsing raw strings into list of int tokens.

    Examples:
        >>> parser = CharParser(['a', 'b', 'c'])
        >>> parser('abc')
        [0, 1, 2]
    """

    def __init__(
        self,
        labels: List[str],
        *,
        unk_id: int = -1,
        blank_id: int = -1,
        do_normalize: bool = True,
        do_lowercase: bool = True,
    ):
        """Creates simple mapping char parser.

        Args:
            labels: List of labels to allocate indexes for. Essentially,
                this is a id to str mapping.
            unk_id: Index to choose for OOV words (default: -1).
            blank_id: Index to filter out from final list of tokens
                (default: -1).
            do_normalize: True if apply normalization step before tokenizing
                (default: True).
            do_lowercase: True if apply lowercasing at normalizing step
                (default: True).
        """

        self._labels = labels
        self._unk_id = unk_id
        self._blank_id = blank_id
        self._do_normalize = do_normalize
        self._do_lowercase = do_lowercase

        self._labels_map = {label: index for index, label in enumerate(labels)}
        self._special_labels = set([label for label in labels if len(label) > 1])

    def __call__(self, text: str) -> Optional[List[int]]:
        if self._do_normalize:
            text = self._normalize(text)
            if text is None:
                return None

        text_tokens = self._tokenize(text)

        return text_tokens

    def _normalize(self, text: str) -> Optional[str]:
        text = text.strip()

        if self._do_lowercase:
            text = text.lower()

        return text

    def _tokenize(self, text: str) -> List[int]:
        tokens = []
        # Split by word for find special labels.
        for word_id, word in enumerate(text.split(' ')):
            if word_id != 0:  # Not first word - so we insert space before.
                tokens.append(self._labels_map.get(' ', self._unk_id))

            if word in self._special_labels:
                tokens.append(self._labels_map[word])
                continue

            for char in word:
                tokens.append(self._labels_map.get(char, self._unk_id))

        # If unk_id == blank_id, OOV tokens are removed.
        tokens = [token for token in tokens if token != self._blank_id]

        return tokens


class ENCharParser(CharParser):
    """Incorporates english-specific parsing logic."""

    PUNCTUATION_TO_REPLACE = frozendict.frozendict({'+': 'plus', '&': 'and', '%': 'percent'})

    def __init__(self, *args, **kwargs):
        """Creates english-specific mapping char parser.

        This class overrides normalizing implementation.

        Args:
            *args: Positional args to pass to `CharParser` constructor.
            **kwargs: Key-value args to pass to `CharParser` constructor.
        """

        super().__init__(*args, **kwargs)

        self._table = self.__make_trans_table()

    def __make_trans_table(self):
        punctuation = string.punctuation

        for char in self.PUNCTUATION_TO_REPLACE:
            punctuation = punctuation.replace(char, '')

        for label in self._labels:
            punctuation = punctuation.replace(label, '')

        table = str.maketrans(punctuation, ' ' * len(punctuation))

        return table

    def _normalize(self, text: str) -> Optional[str]:
        # noinspection PyBroadException
        try:
            text = cleaners.clean_text(
                string=text, table=self._table, punctuation_to_replace=self.PUNCTUATION_TO_REPLACE,
            )
        except Exception:
            return None

        return text


NAME_TO_PARSER = frozendict.frozendict({'base': CharParser, 'en': ENCharParser})


def make_parser(labels: Optional[List[str]] = None, name: str = 'base', **kwargs,) -> CharParser:
    """Creates parser from labels, set of arguments and concise parser name.

    Args:
        labels: List of labels to allocate indexes for. If set to
            None then labels would be ascii table list. Essentially, this is a
            id to str mapping (default: None).
        name: Concise name of parser to create (default: 'base').
            (default: -1).
        **kwargs: Other set of kwargs to pass to parser constructor.

    Returns:
        Instance of `CharParser`.

    Raises:
        ValueError: For invalid parser name.

    Examples:
        >>> type(make_parser(['a', 'b', 'c'], 'en'))
        ENCharParser
    """

    if name not in NAME_TO_PARSER:
        raise ValueError('Invalid parser name.')

    if labels is None:
        labels = list(string.printable)

    parser_type = NAME_TO_PARSER[name]
    parser = parser_type(labels=labels, **kwargs)

    return parser



# Copyright (c) 2019 NVIDIA Corporation
from typing import Any, Dict, List, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from .jasper import JasperEncoder
from .parts.jasper import init_weights
from nemo.backends.pytorch.nm import TrainableNM
from nemo.core.neural_types import *
from nemo.utils import logging
from nemo.utils.decorators import add_port_docs


class ContextNetEncoder(JasperEncoder):
    """
    ContextNet Encoder creates the pre-processing (prologue), QuartzNet convolution
    block, and the additional pre and post processing layers as described in
    ContextNet (https://arxiv.org/abs/2005.03191)

    Args:
        jasper (list): A list of dictionaries. Each element in the list
            represents the configuration of one Jasper Block. Each element
            should contain::

                {
                    # Required parameters
                    'filters' (int) # Number of output channels,
                    'repeat' (int) # Number of sub-blocks,
                    'kernel' (int) # Size of conv kernel,
                    'stride' (int) # Conv stride
                    'dilation' (int) # Conv dilation
                    'dropout' (float) # Dropout probability
                    'residual' (bool) # Whether to use residual or not.
                    # Optional parameters
                    'residual_dense' (bool) # Whether to use Dense Residuals
                        # or not. 'residual' must be True for 'residual_dense'
                        # to be enabled.
                        # Defaults to False.
                    'separable' (bool) # Whether to use separable convolutions.
                        # Defaults to False
                    'groups' (int) # Number of groups in each conv layer.
                        # Defaults to 1
                    'heads' (int) # Sharing of separable filters
                        # Defaults to -1
                    'tied' (bool)  # Whether to use the same weights for all
                        # sub-blocks.
                        # Defaults to False
                    'se' (bool)  # Whether to add Squeeze and Excitation
                        # sub-blocks.
                        # Defaults to False
                    'se_reduction_ratio' (int)  # The reduction ratio of the Squeeze
                        # sub-module.
                        # Must be an integer > 1.
                        # Defaults to 8.
                    'se_context_window' (int) # The size of the temporal context
                        # provided to SE sub-module.
                        # Must be an integer. If value <= 0, will perform global
                        # temporal pooling (global context).
                        # If value >= 1, will perform stride 1 average pooling to
                        # compute context window.
                    'se_interpolation_mode' (str) # Interpolation mode of timestep dimension.
                        # Used only if context window is > 1.
                        # The modes available for resizing are: `nearest`, `linear` (3D-only),
                        # `bilinear`, `area`
                    'kernel_size_factor' (float)  # Conv kernel size multiplier
                        # Can be either an int or float
                        # Kernel size is recomputed as below:
                        # new_kernel_size = int(max(1, (kernel_size * kernel_width)))
                        # to prevent kernel sizes than 1.
                        # Note: If rescaled kernel size is an even integer,
                        # adds 1 to the rescaled kernel size to allow "same"
                        # padding.
                    'stride_last' (bool) # Bool flag to determine whether each
                        # of the the repeated sub-blockss will perform a stride,
                        # or only the last sub-block will perform a strided convolution.
                }

        activation (str): Activation function used for each sub-blocks. Can be
            one of ["hardtanh", "relu", "selu", "swish"].
        feat_in (int): Number of channels being input to this module
        normalization_mode (str): Normalization to be used in each sub-block.
            Can be one of ["batch", "layer", "instance", "group"]
            Defaults to "batch".
        residual_mode (str): Type of residual connection.
            Can be "add", "stride_add" or "max".
            "stride_add" mode performs strided convolution prior to residual
            addition.
            Defaults to "add".
        norm_groups (int): Number of groups for "group" normalization type.
            If set to -1, number of channels is used.
            Defaults to -1.
        conv_mask (bool): Controls the use of sequence length masking prior
            to convolutions.
            Defaults to True.
        frame_splicing (int): Defaults to 1.
        init_mode (str): Describes how neural network parameters are
            initialized. Options are ['xavier_uniform', 'xavier_normal',
            'kaiming_uniform','kaiming_normal'].
            Defaults to "xavier_uniform".
    """

    length: Optional[torch.Tensor]

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "audio_signal": NeuralType(
            #    {0: AxisType(BatchTag), 1: AxisType(SpectrogramSignalTag), 2: AxisType(ProcessedTimeTag),}
            # ),
            # "length": NeuralType({0: AxisType(BatchTag)}),
            "audio_signal": NeuralType(('B', 'D', 'T'), SpectrogramType()),
            "length": NeuralType(tuple('B'), LengthsType()),
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        return {
            # "outputs": NeuralType(
            #    {0: AxisType(BatchTag), 1: AxisType(EncodedRepresentationTag), 2: AxisType(ProcessedTimeTag),}
            # ),
            # "encoded_lengths": NeuralType({0: AxisType(BatchTag)}),
            "outputs": NeuralType(('B', 'D', 'T'), AcousticEncodedRepresentation()),
            "encoded_lengths": NeuralType(tuple('B'), LengthsType()),
        }

    def __init__(
        self,
        jasper: List[Dict[str, Any]],
        activation: str,
        feat_in: int,
        normalization_mode: str = "batch",
        residual_mode: str = "add",
        norm_groups: int = -1,
        conv_mask: bool = False,
        frame_splicing: int = 1,
        init_mode: str = 'xavier_uniform',
    ):
        super().__init__(
            jasper=jasper,
            activation=activation,
            feat_in=feat_in,
            normalization_mode=normalization_mode,
            residual_mode=residual_mode,
            norm_groups=norm_groups,
            conv_mask=conv_mask,
            frame_splicing=frame_splicing,
            init_mode=init_mode,
        )


class ContextNetDecoderForCTC(TrainableNM):
    """
    ContextNet Decoder creates the final layer in ContextNet that maps from the outputs
    of ContextNet Encoder to the vocabulary of interest.

    Args:
        feat_in (int): Number of channels being input to this module
        num_classes (int): Number of characters in ASR model's vocab/labels.
            This count should not include the CTC blank symbol.
        hidden_size (int): Number of units in the hidden state of the LSTM RNN.
        init_mode (str): Describes how neural network parameters are
            initialized. Options are ['xavier_uniform', 'xavier_normal',
            'kaiming_uniform','kaiming_normal'].
            Defaults to "xavier_uniform".
    """

    @property
    @add_port_docs()
    def input_ports(self):
        """Returns definitions of module input ports.
        """
        return {
            # "encoder_output": NeuralType(
            #    {0: AxisType(BatchTag), 1: AxisType(EncodedRepresentationTag), 2: AxisType(ProcessedTimeTag),}
            # )
            "encoder_output": NeuralType(('B', 'D', 'T'), AcousticEncodedRepresentation())
        }

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.
        """
        # return {"output": NeuralType({0: AxisType(BatchTag), 1: AxisType(TimeTag), 2: AxisType(ChannelTag),})}
        return {"output": NeuralType(('B', 'T', 'D'), LogprobsType())}

    def __init__(self, feat_in: int, num_classes: int, hidden_size: int = 640, init_mode: str = "xavier_uniform"):
        super().__init__()

        self._feat_in = feat_in
        # Add 1 for blank char
        self._num_classes = num_classes + 1

        self.rnn = nn.LSTM(feat_in, hidden_size, bias=True, batch_first=True)
        self.clf = nn.Linear(hidden_size, self._num_classes)
        self.clf.apply(lambda x: init_weights(x, mode=init_mode))
        self.to(self._device)

    def forward(self, encoder_output):
        encoder_output = encoder_output.transpose(1, 2)  # [B, T, D]
        output, states = self.rnn(encoder_output)
        logits = self.clf(output)
        return F.log_softmax(logits, dim=-1)



# ! /usr/bin/python
# -*- coding: utf-8 -*-

# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import copy
import os
from inspect import signature
from collections import OrderedDict
from pathlib import Path
import urllib.request
import numpy as np

# git clone git@github.com:microsoft/onnxruntime.git
# cd onnxruntime
#
# ./build.sh --update --build --config RelWithDebInfo  --build_shared_lib --parallel \
#     --cudnn_home /usr/lib/x86_64-linux-gnu --cuda_home /usr/local/cuda \
#     --tensorrt_home .../TensorRT --use_tensorrt --enable_pybind --build_wheel
#
# pip install --upgrade ./build/Linux/RelWithDebInfo/dist/*.whl
import onnxruntime as ort
import pytest
import torch

import nemo
import nemo.collections.nlp as nemo_nlp
import nemo.collections.nlp.nm.trainables.common.token_classification_nm
import nemo.collections.tts as nemo_tts

from nemo import logging
from nemo.core import DeploymentFormat as DF
from nemo.core import NeuralModule

# Check if the required libraries and runtimes are installed.
# Only initialize GPU after this runner is activated.
__trt_pycuda_req_satisfied = True
try:
    import pycuda.autoinit

    # This import causes pycuda to automatically manage CUDA context creation and cleanup.
    import pycuda.driver as cuda

    from .tensorrt_loaders import (
        DefaultDataLoader,
        DataLoaderCache,
        OnnxFileLoader,
        OnnxNetworkLoader,
        BuildEngineLoader,
    )
    from .tensorrt_runner import TensorRTRunnerV2
except Exception as e:
    nemo.logging.error('Failed to import: `{}` ({})'.format(str(e), type(e)))
    __trt_pycuda_req_satisfied = False

# create decorator so that tests can be marked with the TRT requirement
requires_trt = pytest.mark.skipif(
    not __trt_pycuda_req_satisfied, reason="TensorRT/PyCuda library required to run test"
)


@pytest.mark.usefixtures("neural_factory")
class TestDeployExport:
    @torch.no_grad()
    def __test_export_route(self, module, out_name, mode, input_example=None):
        # select correct extension based on the output format
        ext = {DF.ONNX: ".onnx", DF.TRTONNX: ".trt.onnx", DF.PYTORCH: ".pt", DF.TORCHSCRIPT: ".ts"}.get(mode, ".onnx")
        out = Path(f"{out_name}{ext}")
        out_name = str(out)

        if out.exists():
            os.remove(out)

        module.eval()
        torch.manual_seed(1)
        deploy_input_example = input_example
        if isinstance(input_example, OrderedDict):
            deploy_input_example = tuple(input_example.values())
            if len(deploy_input_example) == 1:
                deploy_input_example = deploy_input_example[0]
        elif isinstance(input_example, tuple):
            deploy_input_example = input_example if len(input_example) > 1 else input_example[0]

        sig = signature(module.forward)
        pnum = len(sig.parameters)
        outputs_fwd = module.forward(*deploy_input_example) if pnum > 2 else module.forward(deploy_input_example)
        self.nf.deployment_export(
            module=module, output=out_name, input_example=deploy_input_example, d_format=mode, output_example=None,
        )

        assert out.exists() == True

        if mode == DF.TRTONNX:

            data_loader = DefaultDataLoader()
            loader_cache = DataLoaderCache(data_loader)
            profile_shapes = OrderedDict()
            names = list(module.input_ports) + list(module.output_ports)
            names = list(
                filter(
                    lambda x: x
                    not in (module._disabled_deployment_input_ports | module._disabled_deployment_output_ports),
                    names,
                )
            )
            if isinstance(input_example, tuple):
                si = [tuple(input_example[i].shape) for i in range(len(input_example))]
            elif isinstance(input_example, OrderedDict):
                si = [tuple(input_example.values())[i].shape for i in range(len(input_example))]
            else:
                si = [tuple(input_example.shape)]
            if isinstance(outputs_fwd, tuple):
                fi = [tuple(outputs_fwd[i].shape) for i in range(len(outputs_fwd))]
            else:
                fi = [tuple(outputs_fwd.shape)]
            si = si + fi
            i = 0
            for name in names:
                profile_shapes[name] = [si[i]] * 3
                i = i + 1

            onnx_loader = OnnxFileLoader(out_name)
            network_loader = OnnxNetworkLoader(onnx_loader, explicit_precision=False)
            model_loader = BuildEngineLoader(
                network_loader,
                max_workspace_size=1 << 30,
                fp16_mode=False,
                int8_mode=False,
                profile_shapes=profile_shapes,
                write_engine=None,
                calibrator=None,
                layerwise=False,
            )

            with TensorRTRunnerV2(model_loader=model_loader) as active_runner:
                input_metadata = active_runner.get_input_metadata()
                if input_metadata is None:
                    logging.critical("For {:}, get_input_metadata() returned None!".format(active_runner.name))
                logging.debug("Runner Inputs: {:}".format(input_metadata))
                feed_dict = loader_cache.load(iteration=0, input_metadata=input_metadata, input_example=input_example)
                inputs = dict()
                input_names = list(input_metadata.keys())
                for i in range(len(input_names)):
                    input_name = input_names[i]
                    if input_name in module._disabled_deployment_input_ports:
                        continue

                    if isinstance(input_example, OrderedDict):
                        for key in input_example.keys():
                            if key in input_name:
                                inputs[input_name] = input_example[key].cpu().numpy()
                    elif isinstance(input_example, tuple):
                        inputs[input_name] = input_example[i].cpu().numpy()
                    else:
                        inputs[input_name] = input_example.cpu().numpy()

                out_dict = active_runner.infer(feed_dict=feed_dict, output=outputs_fwd)
                for ov in out_dict.values():
                    outputs_scr = torch.from_numpy(ov).cuda()
                    break

                outputs = []
                outputs.append(copy.deepcopy(out_dict))
                logging.debug(
                    "Received outputs: {:}".format(
                        ["{:}: {:}".format(name, out.shape) for name, out in out_dict.items()]
                    )
                )
                logging.info("Output Buffers: {:}".format(outputs))

            inpex = []
            for ie in feed_dict.values():  # loader_cache.cache[0].values():
                if ie.dtype.type is np.int32:
                    inpex.append(torch.from_numpy(ie).long().cuda())
                else:
                    inpex.append(torch.from_numpy(ie).cuda())
                if len(inpex) == len(input_example):
                    break
            inpex = tuple(inpex)
            outputs_fwd = module.forward(*inpex)

        elif mode == DF.ONNX:
            # Must recompute because *module* might be different now
            torch.manual_seed(1)
            outputs_fwd = module.forward(*deploy_input_example) if pnum > 2 else module.forward(deploy_input_example)

            sess_options = ort.SessionOptions()
            sess_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_BASIC
            ort_session = ort.InferenceSession(out_name, sess_options, ['CUDAExecutionProvider'])
            print('Execution Providers: ', ort_session.get_providers())
            inputs = dict()
            input_names = (
                list(input_example.keys())
                if isinstance(input_example, OrderedDict)
                else list(module.input_ports.keys())
            )
            ort_inputs = ort_session.get_inputs()

            for node_arg in ort_inputs:
                ort_name = node_arg.name
                for input_name in input_names:
                    if input_name in ort_name or ort_name in input_name:
                        break
                if ort_name not in inputs:
                    inputs[ort_name] = (
                        input_example[input_name].cpu().numpy()
                        if isinstance(input_example, OrderedDict)
                        else input_example.cpu().numpy()
                    )

            output_names = None
            outputs_scr = ort_session.run(output_names, inputs)
            outputs_scr = torch.from_numpy(outputs_scr[0]).cuda()
        elif mode == DF.TORCHSCRIPT:
            tscr = torch.jit.load(out_name)
            torch.manual_seed(1)
            outputs_scr = (
                tscr.forward(*tuple(input_example.values()))
                if isinstance(input_example, OrderedDict)
                else (
                    tscr.forward(*input_example) if isinstance(input_example, tuple) else tscr.forward(input_example)
                )
            )
        elif mode == DF.PYTORCH:
            module.restore_from(out_name)
            torch.manual_seed(1)
            if isinstance(input_example, OrderedDict):
                outputs_scr = module.forward(*tuple(input_example.values()))
            elif isinstance(input_example, tuple) or isinstance(input_example, list):
                outputs_scr = module.forward(*input_example)
            else:
                outputs_scr = module.forward(input_example)

        outputs_scr = (
            outputs_scr[0] if isinstance(outputs_scr, tuple) or isinstance(outputs_scr, list) else outputs_scr
        )
        outputs_fwd = (
            outputs_fwd[0] if isinstance(outputs_fwd, tuple) or isinstance(outputs_fwd, list) else outputs_fwd
        )

        n = outputs_fwd.numel()
        tol = 5.0e-3 if n < 10000 else (5.0e-2 if n < 100000 else (5.0e-1))

        assert (outputs_scr - outputs_fwd).norm(p=2) < tol

        if out.exists():
            os.remove(out)

    @pytest.mark.unit
    @pytest.mark.run_only_on('GPU')
    @pytest.mark.parametrize(
        "input_example, module_name, df_type",
        [
            # TaylorNet export tests.
            (torch.randn(4, 1), "TaylorNet", DF.PYTORCH),
            # TokenClassifier export tests.
            (torch.randn(16, 16, 512), "TokenClassifier", DF.ONNX),
            (torch.randn(16, 16, 512), "TokenClassifier", DF.TORCHSCRIPT),
            (torch.randn(16, 16, 512), "TokenClassifier", DF.PYTORCH),
            pytest.param(torch.randn(16, 16, 512), "TokenClassifier", DF.TRTONNX, marks=requires_trt),
            # JasperDecoderForCTC export tests.
            (torch.randn(34, 1024, 1), "JasperDecoderForCTC", DF.ONNX),
            (torch.randn(34, 1024, 1), "JasperDecoderForCTC", DF.TORCHSCRIPT),
            (torch.randn(34, 1024, 1), "JasperDecoderForCTC", DF.PYTORCH),
            pytest.param(torch.randn(34, 1024, 1), "JasperDecoderForCTC", DF.TRTONNX, marks=requires_trt),
            # JasperEncoder export tests.
            (torch.randn(16, 64, 256), "JasperEncoder", DF.ONNX),
            (torch.randn(16, 64, 256), "JasperEncoder", DF.TORCHSCRIPT),
            (torch.randn(16, 64, 256), "JasperEncoder", DF.PYTORCH),
            pytest.param(torch.randn(16, 64, 256), "JasperEncoder", DF.TRTONNX, marks=requires_trt),
            # QuartznetEncoder export tests.
            (torch.randn(16, 64, 256), "QuartznetEncoder", DF.ONNX),
            (torch.randn(16, 64, 256), "QuartznetEncoder", DF.TORCHSCRIPT),
            (torch.randn(16, 64, 256), "QuartznetEncoder", DF.PYTORCH),
            pytest.param(torch.randn(16, 64, 256), "QuartznetEncoder", DF.TRTONNX, marks=requires_trt),
        ],
    )
    def test_module_export(self, tmpdir, input_example, module_name, df_type):
        """ Tests the module export.

            Args:
                tmpdir: Fixture which will provide a temporary directory.

                input_example: Input to be passed to TaylorNet.

                module_name: Name of the module (section in config file).

                df_type: Parameter denoting type of export to be tested.
        """
        # Create neural module instance.
        module = NeuralModule.import_from_config("tests/configs/test_deploy_export.yaml", module_name)
        # Generate filename in the temporary directory.
        tmp_file_name = str(tmpdir.mkdir("export").join(module_name))
        input_example = input_example.cuda() if input_example is not None else input_example
        # Test export.
        self.__test_export_route(
            module=module, out_name=tmp_file_name, mode=df_type, input_example=input_example,
        )

    @pytest.mark.unit
    @pytest.mark.run_only_on('GPU')
    @pytest.mark.parametrize("df_type", [DF.ONNX, DF.TORCHSCRIPT, DF.PYTORCH])
    def test_hf_bert(self, tmpdir, df_type):
        """ Tests BERT export.

            Args:
                tmpdir: Fixture which will provide a temporary directory.

                df_type: Parameter denoting type of export to be tested.
        """
        bert = nemo.collections.nlp.nm.trainables.common.huggingface.BERT(pretrained_model_name="bert-base-uncased")
        input_example = OrderedDict(
            [
                ("input_ids", torch.randint(low=0, high=16, size=(2, 16)).cuda()),
                ("token_type_ids", torch.randint(low=0, high=2, size=(2, 16)).cuda()),
                ("attention_mask", torch.randint(low=0, high=2, size=(2, 16)).cuda()),
            ]
        )
        # Generate filename in the temporary directory.
        tmp_file_name = str(tmpdir.mkdir("export").join("bert"))
        # Test export.
        self.__test_export_route(module=bert, out_name=tmp_file_name, mode=df_type, input_example=input_example)

    @pytest.mark.unit
    @pytest.mark.run_only_on('GPU')
    @pytest.mark.parametrize("df_type", [DF.TORCHSCRIPT, DF.PYTORCH])
    #
    # TODO WaveGlow.infer uses torch.randn which is required to be seeded
    # for deterministic results. It gets translated to ONNX op like this:
    #
    #   %16020 = RandomNormalLike[dtype = 1](%16019)
    #
    # There is no way to seed it, thus to validate ONNX test flow
    # please use torch.ones
    #
    # @pytest.mark.parametrize("df_type", [DF.ONNX, DF.TORCHSCRIPT, DF.PYTORCH])
    #
    def test_waveglow(self, tmpdir, df_type):
        url = "https://api.ngc.nvidia.com/v2/models/nvidia/waveglow_ljspeech/versions/2/files/WaveGlowNM.pt"
        ptfile = "./WaveGlowNM.pt"
        if not Path(ptfile).is_file():
            urllib.request.urlretrieve(url, ptfile)

        module = nemo_tts.WaveGlowInferNM(sample_rate=22050)
        module.restore_from(ptfile)
        module.eval()

        torch.manual_seed(1)
        mel = torch.randn(1, 80, 96).cuda()

        input_example = OrderedDict([("mel_spectrogram", mel)])
        tmp_file_name = str(tmpdir.mkdir("export").join("waveglow"))

        self.__test_export_route(module=module, out_name=tmp_file_name, mode=df_type, input_example=input_example)



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
# Copyright 2019 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

'''
This file contains code artifacts adapted from the original implementation:
https://github.com/google-research/google-research/blob/master/schema_guided_dst/baseline/train_and_predict.py
'''

import torch

from nemo import logging
from nemo.backends.pytorch import LossNM
from nemo.collections.nlp.data.datasets.sgd_dataset.input_example import STATUS_ACTIVE
from nemo.core import ChannelType, LabelsType, LogitsType, NeuralType
from nemo.utils.decorators import add_port_docs

__all__ = ['SGDDialogueStateLossNM']


class SGDDialogueStateLossNM(LossNM):
    """
    Neural module which implements loss for SGD model.
    """

    @property
    @add_port_docs
    def input_ports(self):
        """Returns definitions of module input ports.
            logit_intent_status (float): Output of SGD model
            intent_status_labels (int): Intent labels
            logit_req_slot_status (float): Output of SGD model
            requested_slot_status (float): Takes value 1 if the corresponding slot is requested, 0 otherwise
            req_slot_mask (bool): Masks requested slots not used for the particular service
            logit_cat_slot_status (float): Output of SGD model
            categorical_slot_status (int): The status of each categorical slot in the service
            cat_slot_status_mask (bool): Masks categorical slots not used for the particular service
            logit_cat_slot_value (float): Output of SGD model
            categorical_slot_values (int): The index of the correct value for each categorical slot
            logit_noncat_slot_status (float): Output of SGD model
            noncategorical_slot_status (int): The status of each noncategorical slot in the service
            noncat_slot_status_mask (bool): masks noncategorical slots not used for the particular service
            logit_noncat_slot_start (float): Output of SGD model
            logit_noncat_slot_end (float): Output of SGD model
            noncategorical_slot_value_start (int): The index of the starting subword corresponding to the slot span for a non-categorical slot value
            noncategorical_slot_value_end (int): The index of the ending (inclusive) subword corresponding to the slot span for a non-categorical slot value
        """
        return {
            "logit_intent_status": NeuralType(('B', 'T', 'C'), LogitsType()),
            "intent_status_labels": NeuralType(('B'), LabelsType()),
            "logit_req_slot_status": NeuralType(('B', 'T'), LogitsType()),
            "requested_slot_status": NeuralType(('B', 'T'), LabelsType()),
            "req_slot_mask": NeuralType(('B', 'T'), ChannelType()),
            "logit_cat_slot_status": NeuralType(('B', 'T', 'C'), LogitsType()),
            "categorical_slot_status": NeuralType(('B', 'T'), LabelsType()),
            "cat_slot_status_mask": NeuralType(('B', 'T'), ChannelType()),
            "logit_cat_slot_value": NeuralType(('B', 'T', 'C'), LogitsType()),
            "categorical_slot_values": NeuralType(('B', 'T'), LabelsType()),
            "logit_noncat_slot_status": NeuralType(('B', 'T', 'C'), LogitsType()),
            "noncategorical_slot_status": NeuralType(('B', 'T'), LabelsType()),
            "noncat_slot_status_mask": NeuralType(('B', 'T'), ChannelType()),
            "logit_noncat_slot_start": NeuralType(('B', 'T', 'C'), LogitsType()),
            "logit_noncat_slot_end": NeuralType(('B', 'T', 'C'), LogitsType()),
            "noncategorical_slot_value_start": NeuralType(('B', 'T'), LabelsType()),
            "noncategorical_slot_value_end": NeuralType(('B', 'T'), LabelsType()),
        }

    @property
    def output_ports(self):
        """
        Returns definitions of module output ports.
        loss:
            NeuralType(None)
        """
        return {"loss": NeuralType(None)}

    def __init__(self, add_carry_status, reduction='mean'):
        """
        Args:
            add_carry_status (bool): specifies that carry status is enabled for slots
            reduction (str): specifies the reduction to apply to the final loss, choose 'mean' or 'sum'
        """
        super().__init__()

        self._add_carry_status = add_carry_status

        if reduction not in ['mean', 'sum']:
            logging.warning(f'{reduction} reduction is not supported. Setting reduction to "mean"')
            reduction = 'mean'

        self.reduction = reduction
        self._cross_entropy = torch.nn.CrossEntropyLoss(reduction=self.reduction)
        self._criterion_req_slots = torch.nn.BCEWithLogitsLoss(reduction=self.reduction)

    def _loss_function(
        self,
        logit_intent_status,
        intent_status_labels,
        logit_req_slot_status,
        requested_slot_status,
        req_slot_mask,
        logit_cat_slot_status,
        categorical_slot_status,
        cat_slot_status_mask,
        logit_cat_slot_value,
        categorical_slot_values,
        logit_noncat_slot_status,
        noncategorical_slot_status,
        noncat_slot_status_mask,
        logit_noncat_slot_start,
        logit_noncat_slot_end,
        noncategorical_slot_value_start,
        noncategorical_slot_value_end,
    ):
        # Intent loss
        intent_loss = self._cross_entropy(logit_intent_status, intent_status_labels)

        # Requested slots.
        # Shape: (batch_size, max_num_slots)
        # mask unused slots
        # Sigmoid cross entropy is used because more than one slots can be requested in a single utterance
        req_slot_mask = req_slot_mask > 0.5
        requested_slot_loss = self._criterion_req_slots(
            logit_req_slot_status[req_slot_mask], requested_slot_status[req_slot_mask]
        )

        # Categorical slot status
        # Shape of logit_cat_slot_status: (batch_size, max_num_cat_slots, 3)
        # cat_slot_status_mask masks unused categorical padded slots for the service
        cat_slot_status_mask = cat_slot_status_mask.view(-1) > 0.5
        if sum(cat_slot_status_mask) == 0:
            logging.warning(f'No categorical slots in the batch')
            cat_slot_status_loss = torch.clamp(torch.max(logit_cat_slot_status.view(-1)), 0, 0)
        else:
            cat_slot_status_loss = self._cross_entropy(
                logit_cat_slot_status.view(-1, logit_cat_slot_status.size()[-1])[cat_slot_status_mask],
                categorical_slot_status.view(-1)[cat_slot_status_mask],
            )

        # Categorical slot values.
        # Shape: (batch_size, max_num_cat_slots, max_num_slot_values).
        max_num_slot_values = logit_cat_slot_value.size()[-1]

        # Zero out losses for categorical slot value when the slot status is not active.
        if self._add_carry_status:
            cat_slot_value_mask = (categorical_slot_status >= STATUS_ACTIVE).view(-1)
        else:
            cat_slot_value_mask = (categorical_slot_status == STATUS_ACTIVE).view(-1)

        # to handle cases with no active categorical slot value
        if sum(cat_slot_value_mask) == 0:
            logging.warning(f'No active values for categorical slots in the batch.')
            cat_slot_value_loss = torch.clamp(torch.max(logit_cat_slot_value.view(-1)), 0, 0)
        else:
            slot_values_active_logits = logit_cat_slot_value.view(-1, max_num_slot_values)[cat_slot_value_mask]
            slot_values_active_labels = categorical_slot_values.view(-1)[cat_slot_value_mask]
            cat_slot_value_loss = self._cross_entropy(slot_values_active_logits, slot_values_active_labels)

        # Non-categorical slot status.
        # Shape: (batch_size, max_num_noncat_slots, 3).
        # noncat_slot_status_mask masks unused noncat slots for the service
        noncat_slot_status_mask = noncat_slot_status_mask.view(-1) > 0.5
        if sum(noncat_slot_status_mask) == 0:
            logging.warning(f'No active non-categorical slots in the batch.')
            noncat_slot_status_loss = torch.clamp(torch.max(logit_noncat_slot_status.view(-1)), 0, 0)
        else:
            noncat_slot_status_loss = self._cross_entropy(
                logit_noncat_slot_status.view(-1, logit_noncat_slot_status.size()[-1])[noncat_slot_status_mask],
                noncategorical_slot_status.view(-1)[noncat_slot_status_mask],
            )

        # Non-categorical slot spans.
        # Shape: (batch_size, max_num_noncat_slots, max_num_tokens).n
        max_num_tokens = logit_noncat_slot_start.size()[-1]
        # Zero out losses for non-categorical slot spans when the slot status is not active.
        if self._add_carry_status:
            non_cat_slot_value_mask = (noncategorical_slot_status >= STATUS_ACTIVE).view(-1)
        else:
            non_cat_slot_value_mask = (noncategorical_slot_status == STATUS_ACTIVE).view(-1)

        # to handle cases with no active categorical slot value
        if sum(non_cat_slot_value_mask) == 0:
            logging.warning(f'No active values for non-categorical slots in the batch.')
            span_start_loss = torch.clamp(torch.max(logit_noncat_slot_start.view(-1)), 0, 0)
            span_end_loss = torch.clamp(torch.max(logit_noncat_slot_end.view(-1)), 0, 0)
        else:
            noncat_slot_start_active_logits = logit_noncat_slot_start.view(-1, max_num_tokens)[non_cat_slot_value_mask]
            noncat_slot_start_active_labels = noncategorical_slot_value_start.view(-1)[non_cat_slot_value_mask]
            span_start_loss = self._cross_entropy(noncat_slot_start_active_logits, noncat_slot_start_active_labels)

            noncat_slot_end_active_logits = logit_noncat_slot_end.view(-1, max_num_tokens)[non_cat_slot_value_mask]
            noncat_slot_end_active_labels = noncategorical_slot_value_end.view(-1)[non_cat_slot_value_mask]
            span_end_loss = self._cross_entropy(noncat_slot_end_active_logits, noncat_slot_end_active_labels)

        losses = {
            "intent_loss": intent_loss,
            "requested_slot_loss": requested_slot_loss,
            "cat_slot_status_loss": cat_slot_status_loss,
            "cat_slot_value_loss": cat_slot_value_loss,
            "noncat_slot_status_loss": noncat_slot_status_loss,
            "span_start_loss": span_start_loss,
            "span_end_loss": span_end_loss,
        }

        total_loss = sum(losses.values())
        if self.reduction == 'mean':
            total_loss = total_loss / len(losses)
        else:
            batch_size = logit_intent_status.shape[0]
            total_loss = total_loss / batch_size
        return total_loss



# Copyright (C) NVIDIA CORPORATION. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.****

import threading


class Singleton(type):
    """ Implementation of a generic, tread-safe singleton meta-class.
        Can be used as meta-class, i.e. will create 
    """

    # List of instances - one per class.
    __instances = {}
    # Lock used for accessing the instance.
    __lock = threading.Lock()

    def __call__(cls, *args, **kwargs):
        """ Returns singleton instance. A thread safe implementation. """
        if cls not in cls.__instances:
            # Enter critical section.
            with cls.__lock:
                # Check once again.
                if cls not in cls.__instances:
                    # Create a new object instance - one per class.
                    cls.__instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        # Return the instance.
        return cls.__instances[cls]



# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Any, Dict, List, Optional

import numpy as np
import pandas as pd
import torch
from torch import nn
from torch.nn import functional as F

import nemo
from nemo.backends.pytorch import nm as nemo_nm
from nemo.backends.pytorch.nm import DataLayerNM, LossNM, NonTrainableNM
from nemo.collections import asr as nemo_asr
from nemo.collections import tts as nemo_tts
from nemo.collections.asr.parts import AudioDataset, WaveformFeaturizer
from nemo.core.neural_types import (
    AudioSignal,
    ChannelType,
    EmbeddedTextType,
    EncodedRepresentation,
    LengthsType,
    MaskType,
    MelSpectrogramType,
    NeuralType,
)
from nemo.utils.decorators import add_port_docs

__all__ = [
    'TalkNetDataLayer',
    'TalkNet',
    'LenSampler',
    'TalkNetDursLoss',
    'TalkNetMelsLoss',
]


class Ops:
    """Bunch of PyTorch useful function for tensors manipulation."""

    @staticmethod
    def merge(tensors, value=0, dtype=None):
        tensors = [tensor if isinstance(tensor, torch.Tensor) else torch.tensor(tensor) for tensor in tensors]
        dtype = tensors[0].dtype if dtype is None else dtype
        max_len = max(tensor.shape[0] for tensor in tensors)
        new_tensors = []
        for tensor in tensors:
            pad = (2 * len(tensor.shape)) * [0]
            pad[-1] = max_len - tensor.shape[0]
            new_tensors.append(F.pad(tensor, pad=pad, value=value))
        return torch.stack(new_tensors).to(dtype=dtype)

    @staticmethod
    def make_mask(lengths):
        device = lengths.device if torch.is_tensor(lengths) else 'cpu'
        return Ops.merge([torch.ones(length, device=device) for length in lengths], value=0, dtype=torch.bool)

    @staticmethod
    def interleave(x, y):
        xy = torch.stack([x[:-1], y], dim=1).view(-1)
        xy = F.pad(xy, pad=[0, 1], value=x[-1])
        return xy

    @staticmethod
    def pad(x, to=16, dim=1, value=0):
        pads = [0, 0] * len(x.shape)
        pads[-(2 * dim + 1)] = (to - (x.shape[dim] % to)) % to
        return F.pad(x, pads, value=value)


class TalkNetDataset:
    def __init__(
        self, audio_dataset, durs_file, durs_type='full-pad', speakers=None, speaker_table=None, speaker_embs=None,
    ):
        """TalkNet dataset with indexing.

        Args:
            audio_dataset: Instance of AudioDataset with basic audio/text info.
            durs_file: Durations list file.
            durs_type: Type of durations to use.
            speakers: Speakers list file.
            speaker_table: Table of speakers ids.
            speaker_embs: Matrix of speakers embeddings.
        """

        self._audio_dataset = audio_dataset
        self._durs = np.load(durs_file, allow_pickle=True)
        self._durs_type = durs_type

        self._speakers, self._speakers_table, self._speaker_embs = None, None, None

        if speakers is not None:
            self._speakers = np.load(speakers, allow_pickle=True)

        if speaker_table is not None:
            self._speakers_table = {sid: i for i, sid in enumerate(pd.read_csv(speaker_table, sep='\t').index)}

        if speaker_embs is not None:
            self._speaker_embs = np.load(speaker_embs, allow_pickle=True)

    def __getitem__(self, index):
        (audio, audio_len, text, text_len), misc = self._audio_dataset[index]
        id_, text_raw, speaker = misc['id'], misc['text_raw'], misc['speaker']
        example = dict(audio=audio, audio_len=audio_len, text=text, text_len=text_len, text_raw=text_raw)

        if self._durs_type == 'pad':
            dur = self._durs[id_]
            example['dur'] = torch.tensor(dur, dtype=torch.long)
        elif self._durs_type == 'full-pad':
            blank, dur = self._durs[id_]
            example['blank'] = torch.tensor(blank, dtype=torch.long)
            example['dur'] = torch.tensor(dur, dtype=torch.long)
        else:
            raise ValueError("Wrong durations handling type.")

        if self._speakers_table is not None:
            example['speaker'] = self._speakers_table[speaker]

        if self._speakers is not None:
            example['speaker_emb'] = torch.tensor(self._speakers[id_])

        if self._speaker_embs is not None:
            example['speaker_emb'] = torch.tensor(self._speaker_embs[example['speaker']])

        return example

    def __len__(self):
        return len(self._audio_dataset)


class LengthsAwareSampler(torch.utils.data.distributed.DistributedSampler):  # noqa
    def __init__(self, *args, **kwargs):
        """Assignees samples with similar lengths to reduce padding.

        Args:
            *args: Args to propagate to `DistributedSampler` constructor.
            **kwargs: Kwargs to propagate to `DistributedSampler` constructor. Additional
                keys should be 'lengths' and 'batch_size'.
        """

        self.lengths = kwargs.pop('lengths')
        self.batch_size = kwargs.pop('batch_size')

        super().__init__(*args, **kwargs)

    @staticmethod
    def _local_shuffle(values, keys, window, seed):
        g = np.random.Generator(np.random.PCG64(seed))

        keys = np.array(keys)
        kam_middle = keys[2 * window :] - keys[: -(2 * window)]
        kam_prefix = keys[window : 2 * window] - keys[:window]
        kam_suffix = keys[-window:] - keys[-2 * window : -window]
        kam = np.hstack([kam_prefix, kam_middle, kam_suffix])
        new_keys = keys + g.uniform(-kam, kam, size=keys.shape)

        kvs = list(zip(new_keys, values))
        kvs.sort()

        return [v for _, v in kvs]

    def __iter__(self):
        indices = list(super().__iter__())

        indices.sort(key=lambda i: self.lengths[i])

        # Local shuffle (to introduce a little bit of randomness)
        lens = [self.lengths[i] for i in indices]
        # For lengths over workers to be notably different, coef should be around 1000. But smaller values works for
        # shuffling as well.
        new_indices = self._local_shuffle(indices, lens, 10 * self.batch_size, hash((self.rank, self.epoch)))
        assert len(indices) == len(new_indices)
        indices = new_indices

        batches = []
        for i in range(0, len(indices), self.batch_size):
            batches.append(indices[i : i + self.batch_size])

        g = torch.Generator()
        g.manual_seed(self.epoch)  # noqa
        b_indices = torch.randperm(len(batches), generator=g).tolist()

        for b_i in b_indices:
            yield from batches[b_i]


class AllSampler(torch.utils.data.distributed.DistributedSampler):  # noqa
    def __iter__(self):
        return iter(list(range(len(self.dataset))))


class BlanksDurationAugmentation:
    """Different blanks/durs augs."""

    @staticmethod
    def _split(d):
        return np.array(d[::2]), np.array(d[1::2])

    @staticmethod
    def _merge(b, d):
        result = []
        for b1, d1 in zip(b, d):
            result.extend([b1, d1])

        result.append(b[-1])
        return np.array(result)

    @classmethod
    def shake_biased(cls, b, d, p=0.1):
        """Biased binomial shake."""

        b, d, total = b.copy(), d.copy(), sum(b) + sum(d)

        def split2(x):
            xl = np.random.binomial(x, p, size=x.shape)
            return xl, x - xl

        def split3(x):
            xl, xm = split2(x)
            xr, xm = split2(xm)
            return xl, xm, xr

        bdl, bdm, bdr = split3(cls._merge(b, d))
        bd = bdm + np.roll(bdl, -1) + np.roll(bdr, +1)
        b, d = cls._split(bd)

        assert sum(b) + sum(d) == total

        return b, d

    @classmethod
    def shake_unbiased(cls, b, d, p=0.1):
        b, d, total = b.copy(), d.copy(), sum(b) + sum(d)

        def split2(x, mm):
            xl = np.random.binomial(np.minimum(x, mm), p, size=x.shape)
            return xl, x - xl

        def split3(x):
            xl, xm = split2(x, np.roll(x, +1))
            xr, xm = split2(xm, np.roll(x, -1))
            return xl, xm, xr

        bdl, bdm, bdr = split3(cls._merge(b, d))
        bd = bdm + np.roll(bdl, -1) + np.roll(bdr, +1)
        b, d = cls._split(bd)

        assert sum(b) + sum(d) == total

        return b, d

    @staticmethod
    def zero_out(b, d, p=0.1):
        """Zeros out some of the durs."""

        b, d, total = b.copy(), d.copy(), sum(b) + sum(d)

        mask = np.random.binomial(1, size=d.size, p=p).astype(bool)
        b[:-1][mask] += d[mask]
        d[mask] -= d[mask]

        assert sum(b) + sum(d) == total

        return b, d

    @staticmethod
    def compose(augs):
        """Composes augs."""

        def pipe(b, d, p=0.1):
            for aug in augs:
                b, d = aug(b, d, p=p)

            return b, d

        return pipe


class TalkNetDataLayer(DataLayerNM):
    """Data Layer for TalkNet model.

    Basically, replicated behavior from AudioToText Data Layer, zipped with ground truth durations for additional loss.

    """

    @property
    @add_port_docs
    def output_ports(self):
        """Returns definitions of module output ports."""
        return dict(
            audio=NeuralType(('B', 'T'), AudioSignal(freq=self._sample_rate)),
            audio_len=NeuralType(tuple('B'), LengthsType()),
            text=NeuralType(('B', 'T'), EmbeddedTextType()),
            text_mask=NeuralType(('B', 'T'), MaskType()),
            dur=NeuralType(('B', 'T'), LengthsType()),
            text_rep=NeuralType(('B', 'T'), LengthsType()),
            text_rep_mask=NeuralType(('B', 'T'), MaskType()),
            text_raw=NeuralType(),
            speaker=NeuralType(('B',), EmbeddedTextType(), optional=True),
            speaker_emb=NeuralType(('B', 'T'), EncodedRepresentation(), optional=True),
        )

    def __init__(
        self,
        data: str,
        durs: str,
        labels: List[str],
        durs_type: str = 'full-pad',
        speakers: str = None,
        speaker_table: str = None,
        speaker_embs: str = None,
        batch_size: int = 32,
        sample_rate: int = 16000,
        int_values: bool = False,
        bos_id: Optional[int] = None,
        eos_id: Optional[int] = None,
        pad_id: Optional[int] = None,
        blank_id: Optional[int] = None,
        min_duration: Optional[float] = 0.1,
        max_duration: Optional[float] = None,
        normalize_transcripts: bool = True,
        trim_silence: bool = False,
        load_audio: bool = True,
        drop_last: bool = False,
        shuffle: bool = True,
        num_workers: int = 0,
        sampler_type: str = 'default',
        bd_aug: bool = False,
    ):
        """Creates TalkNet data iterator.

        Args:
            data: Path to dataset manifest file.
            durs: Path to pickled durations file.
            labels: List strings of labels to use.
            durs_type: String id of durations type to use.
            speakers: Speakers list file.
            speaker_table: Speakers ids mapping.
            speaker_embs: Speakers embeddings file.
            batch_size: Number of sample in batch.
            sample_rate: Target sampling rate for data. Audio files will be resampled to sample_rate if it is not
                already.
            int_values: Bool indicating whether the audio file is saved as int data or float data.
            bos_id: Beginning of string symbol id used for seq2seq models.
            eos_id: End of string symbol id used for seq2seq models.
            pad_id: Token used to pad when collating samples in batches.
            blank_id: Int id of blank symbol.
            min_duration: All training files which have a duration less than min_duration are dropped.
            max_duration: All training files which have a duration more than max_duration are dropped.
            normalize_transcripts: Whether to use automatic text cleaning.
            trim_silence: Whether to use trim silence from beginning and end of audio signal using
                librosa.effects.trim().
            load_audio: Controls whether the dataloader loads the audio signal and transcript or just the transcript.
            drop_last: See PyTorch DataLoader.
            shuffle: See PyTorch DataLoader.
            num_workers: See PyTorch DataLoader.
            sampler_type: String id of sampler type to use.
            bd_aug: True if use augmentation for blanks/durs.
        """

        super().__init__()

        # Set up dataset.
        self._featurizer = WaveformFeaturizer(sample_rate=sample_rate, int_values=int_values, augmentor=None)
        dataset_params = {
            'manifest_filepath': data,
            'labels': labels,
            'featurizer': self._featurizer,
            'max_duration': max_duration,
            'min_duration': min_duration,
            'normalize': normalize_transcripts,
            'trim': trim_silence,
            'bos_id': bos_id,
            'eos_id': eos_id,
            'load_audio': load_audio,
            'add_misc': True,
        }
        audio_dataset = AudioDataset(**dataset_params)
        self._dataset = TalkNetDataset(audio_dataset, durs, durs_type, speakers, speaker_table, speaker_embs)
        self._durs_type = durs_type
        self._pad_id = pad_id
        self._blank_id = blank_id
        self._space_id = labels.index(' ')
        self._sample_rate = sample_rate
        self._load_audio = load_audio
        self._bd_aug = bd_aug

        sampler = None
        if self._placement == nemo.core.DeviceType.AllGpu:
            if sampler_type == 'all':
                sampler = AllSampler(self._dataset)
            elif sampler_type == 'default':
                sampler = torch.utils.data.distributed.DistributedSampler(self._dataset)  # noqa
            elif sampler_type == 'super-smart':
                sampler = LengthsAwareSampler(
                    dataset=self._dataset,
                    lengths=[e.duration for e in audio_dataset.collection],
                    batch_size=batch_size,
                )
            else:
                raise ValueError("Invalid sample type.")

        self._dataloader = torch.utils.data.DataLoader(  # noqa
            dataset=self._dataset,
            batch_size=batch_size,
            collate_fn=self._collate,
            drop_last=drop_last,
            shuffle=shuffle if sampler is None else False,
            sampler=sampler,
            num_workers=num_workers,
        )

    def _collate(self, batch):
        batch = {key: [example[key] for example in batch] for key in batch[0]}

        if self._load_audio:
            audio = Ops.merge(batch['audio'])
            audio_len = torch.tensor(batch['audio_len'], dtype=torch.long)
        else:
            audio, audio_len = None, None

        if self._durs_type == 'pad':
            text = [F.pad(text, pad=[1, 1], value=self._space_id) for text in batch['text']]
            text = Ops.merge(text, value=self._pad_id, dtype=torch.long)
            # noinspection PyTypeChecker
            text_mask = Ops.make_mask([text_len + 2 for text_len in batch['text_len']])
            dur = Ops.merge(batch['dur'], dtype=torch.long)
        elif self._durs_type == 'full-pad':
            # `text`
            text = [
                Ops.interleave(x=torch.empty(len(text) + 1, dtype=torch.long).fill_(self._blank_id), y=text)
                for text in batch['text']
            ]
            text = Ops.merge(text, value=self._pad_id, dtype=torch.long)

            # `text_mask`
            text_mask = Ops.make_mask([text_len * 2 + 1 for text_len in batch['text_len']])  # noqa

            # `dur`
            blank, dur = batch['blank'], batch['dur']

            # Aug
            if self._bd_aug:
                if isinstance(self._bd_aug, str):
                    name, p = self._bd_aug.split('|')
                else:
                    name, p = 'shake_biased', '0.1'
                aug = lambda b, d: getattr(BlanksDurationAugmentation, name)(b, d, p)  # noqa

                new_blank, new_dur = [], []
                for b, d in zip(blank, dur):
                    new_b, new_d = aug(b.numpy(), d.numpy())
                    new_blank.append(torch.tensor(new_b))
                    new_dur.append(torch.tensor(new_d))
                blank, dur = new_blank, new_dur

            dur = Ops.merge([Ops.interleave(b, d) for b, d in zip(blank, dur)], dtype=torch.long)
        else:
            raise ValueError("Wrong durations handling type.")

        text_rep = Ops.merge(
            tensors=[torch.repeat_interleave(text1, dur1) for text1, dur1 in zip(text, dur)], dtype=torch.long,
        )
        text_rep_mask = Ops.make_mask(dur.sum(-1))

        text_raw = batch['text_raw']

        speaker, speaker_emb = None, None
        if 'speaker' in batch:
            speaker = torch.tensor(batch['speaker'], dtype=torch.long)
        if 'speaker_emb' in batch:
            speaker_emb = Ops.merge(batch['speaker_emb'], dtype=torch.float)

        assert audio is None or audio.shape[-1] == audio_len.max()
        assert text.shape == text_mask.shape, f'{text.shape} vs {text_mask.shape}'
        assert text.shape == dur.shape, f'{text.shape} vs {dur.shape}'

        return audio, audio_len, text, text_mask, dur, text_rep, text_rep_mask, text_raw, speaker, speaker_emb

    def __len__(self) -> int:
        return len(self._dataset)

    @property
    def dataset(self) -> Optional[torch.utils.data.Dataset]:  # noqa
        return None

    @property
    def data_iterator(self) -> Optional[torch.utils.data.DataLoader]:  # noqa
        return self._dataloader


class PolySpanEmb(nn.Module):
    def __init__(self, emb):
        """Wraps chars embeddings with assigning polynomial span embs for blanks.

        Args:
            emb: Character embeddings.
        """

        super().__init__()

        self._emb = emb

    def forward(self, text, dur):
        lefts, rights = self._generate_sides(text)
        lefts = self._emb(self._generate_text_rep(lefts, dur))
        rights = self._emb(self._generate_text_rep(rights, dur))

        left_c = self._generate_left_c(dur).unsqueeze_(-1)  # noqa

        x = left_c * lefts + (1 - left_c) * rights  # noqa

        return x

    def _generate_sides(self, text):
        lefts = F.pad(text[:, :-1], [1, 0, 0, 0], value=self._emb.padding_idx)
        lefts[:, 1::2] = text[:, 1::2]
        rights = F.pad(text[:, 1:], [0, 1, 0, 0], value=self._emb.padding_idx)
        rights[:, 1::2] = text[:, 1::2]

        return lefts, rights

    @staticmethod
    def _generate_text_rep(text, dur):
        text_rep = []
        for t, d in zip(text, dur):
            text_rep.append(torch.repeat_interleave(t, d))

        text_rep = Ops.merge(text_rep)

        return text_rep

    def _generate_left_c(self, dur):
        x = F.pad(torch.cumsum(dur, dim=-1)[:, :-1], [1, 0], value=0)
        pos_cm = self._generate_text_rep(x, dur)
        mask = self._generate_text_rep(torch.ones_like(dur), dur)
        ones_cm = torch.cumsum(mask, dim=1)
        totals = self._generate_text_rep(dur, dur) + 1

        left_c = (ones_cm - pos_cm) * mask
        left_c = left_c.to(dtype=self._emb.weight.dtype)
        left_c = 1.0 - left_c / totals  # noqa

        return left_c


class TalkNet(nemo_nm.TrainableNM):
    """TalkNet TTS Model"""

    @property
    @add_port_docs
    def input_ports(self):
        """Returns definitions of module input ports."""
        return dict(
            text=NeuralType(('B', 'T'), EmbeddedTextType(), optional=True),
            text_mask=NeuralType(('B', 'T'), MaskType(), optional=True),
            text_rep=NeuralType(('B', 'T'), LengthsType(), optional=True),
            text_rep_mask=NeuralType(('B', 'T'), MaskType(), optional=True),
            speaker_emb=NeuralType(('B', 'T'), EncodedRepresentation(), optional=True),
            durs=NeuralType(('B', 'T'), LengthsType(), optional=True),
        )

    @property
    @add_port_docs
    def output_ports(self):
        """Returns definitions of module output ports."""
        return dict(pred=NeuralType(('B', 'T', 'D'), ChannelType()), len=NeuralType(('B',), LengthsType()))

    def __init__(
        self,
        n_vocab: int,
        d_char: int,
        pad_id: int,
        jasper_kwargs: Dict[str, Any],
        d_out: int,
        d_speaker_emb: Optional[int] = None,
        d_speaker_x: Optional[int] = None,
        d_speaker_o: Optional[int] = None,
        pad16: bool = False,
        poly_span: bool = False,
        doubling: bool = False,
    ):
        """Creates TalkNet backbone instance.

        Args:
            n_vocab: Size of input vocabulary.
            d_char: Dimension of char embedding.
            pad_id: Id of padding symbol.
            jasper_kwargs: Kwargs to instantiate QN encoder.
            d_out: Dimension of output.
            d_speaker_emb: Dimension of speaker embedding.
            d_speaker_x: Dimension of pre speaker embedding.
            d_speaker_o: Dimension of post speaker embedding.
            pad16: True if pad tensors to 16.
            poly_span: True if assign polynomial span embeddings for blanks.
            doubling: True if using mel channels doubling trick.
        """

        super().__init__()

        # Embedding for input text
        self.text_emb = nn.Embedding(n_vocab, d_char, padding_idx=pad_id).to(self._device)
        self.text_emb.weight.data.uniform_(-1, 1)

        # PolySpan
        self.ps = PolySpanEmb(self.text_emb)
        self._poly_span = poly_span

        # Embedding for speaker
        if d_speaker_emb is not None:
            self.speaker_in = nn.Linear(d_speaker_emb, d_speaker_x).to(self._device)
            self.speaker_out = nn.Linear(d_speaker_emb, d_speaker_o).to(self._device)
        else:
            self.speaker_in, self.speaker_out = None, None

        jasper_params = jasper_kwargs['jasper']
        d_enc_out = jasper_params[-1]["filters"]
        d_x = d_char + (int(d_speaker_x or 0) if d_speaker_emb else 0)
        self.jasper = nemo_asr.JasperEncoder(feat_in=d_x, **jasper_kwargs).to(self._device)

        d_o = d_enc_out + (int(d_speaker_o or 0) if d_speaker_emb else 0)
        self.out = nn.Conv1d(d_o, d_out * (1 + int(doubling)), kernel_size=1, bias=True).to(self._device)

        self._pad16 = pad16
        self._doubling = doubling

    def forward(self, text=None, text_mask=None, text_rep=None, text_rep_mask=None, speaker_emb=None, durs=None):
        if self._poly_span:
            # Do not support pad16 yet. Do we really need it though?
            x, x_len = self.ps(text, durs), durs.sum(-1)
        else:
            if text_rep is not None:
                text, text_mask = text_rep, text_rep_mask

            if self._pad16:
                text = Ops.pad(text, value=self.text_emb.padding_idx)

            x, x_len = self.text_emb(text), text_mask.sum(-1)

        if speaker_emb is not None:
            speaker_x = speaker_emb
            if self.speaker_in is not None:
                speaker_x = self.speaker_in(speaker_x)

            speaker_x = speaker_x.unsqueeze(1).repeat([1, x.shape[1], 1])  # BS => BTS
            x = torch.cat([x, speaker_x], dim=-1)  # stack([BTE, BTS]) = BT(E + S)

        o, o_len = self.jasper(x.transpose(-1, -2), x_len, force_pt=True)
        assert x.shape[1] == o.shape[-1]  # Time
        assert torch.equal(x_len, o_len)

        if speaker_emb is not None:
            speaker_o = speaker_emb
            if self.speaker_out is not None:
                speaker_o = self.speaker_out(speaker_o)

            speaker_o = speaker_o.unsqueeze(-1).repeat([1, 1, o.shape[-1]])  # BS => BST
            o = torch.cat([o, speaker_o], dim=1)  # stack([BOT, BST]) = B(O + S)T

        o = o.to(dtype=self.out.weight.dtype)
        o = self.out(o).transpose(-1, -2)  # BTO

        if self._doubling:
            o = o.reshape(o.shape[0], -1, o.shape[-1] // 2)  # Time doubles.

        return o, o_len


class LenSampler(NonTrainableNM):
    """Subsample data based on length to fit in bigger batches."""

    @property
    @add_port_docs
    def input_ports(self):
        """Returns definitions of module input ports."""
        return dict(
            text_rep=NeuralType(('B', 'T'), LengthsType()),
            text_rep_mask=NeuralType(('B', 'T'), MaskType()),
            mel_true=NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            mel_len=NeuralType(('B',), LengthsType()),
        )

    @property
    @add_port_docs
    def output_ports(self):
        """Returns definitions of module input ports."""
        return dict(
            text_rep=NeuralType(('B', 'T'), LengthsType()),
            text_rep_mask=NeuralType(('B', 'T'), MaskType()),
            mel_true=NeuralType(('B', 'D', 'T'), MelSpectrogramType()),
            mel_len=NeuralType(('B',), LengthsType()),
        )

    def __init__(self, max_len=1200):
        super().__init__()

        self._max_len = max_len

    def forward(self, text_rep, text_rep_mask, mel_true, mel_len):
        # Optimization
        if mel_len.max().item() <= self._max_len:
            return text_rep, text_rep_mask, mel_true, mel_len

        inds = np.random.randint(np.maximum(mel_len.cpu().numpy() - self._max_len, 0) + 1)

        text_rep = torch.stack([b1[i : i + self._max_len] for i, b1 in zip(inds, text_rep)])
        text_rep_mask = torch.stack([b1[i : i + self._max_len] for i, b1 in zip(inds, text_rep_mask)])
        mel_true = torch.stack([b1[:, i : i + self._max_len] for i, b1 in zip(inds, mel_true)])
        mel_len = torch.clamp(mel_len, max=self._max_len)

        return text_rep, text_rep_mask, mel_true, mel_len


class TalkNetDursLoss(LossNM):
    """Neural Module Wrapper for TalkNet dur loss."""

    @property
    @add_port_docs
    def input_ports(self):
        """Returns definitions of module input ports."""
        return dict(
            dur_true=NeuralType(('B', 'T'), LengthsType()),
            dur_pred=NeuralType(('B', 'T', 'D'), ChannelType()),
            text_mask=NeuralType(('B', 'T'), MaskType()),
        )

    @property
    @add_port_docs
    def output_ports(self):
        """Returns definitions of module output ports."""
        return dict(loss=NeuralType(None))

    def __init__(
        self,
        method='l2-log',
        num_classes=32,
        dmld_hidden=5,
        reduction='all',
        max_dur=500,
        xe_steps_coef=1.5,
        pad16=False,
    ):
        """Creates duration loss instance.

        Args:
            method: Method for duration loss calculation.
            num_classes: Number of classes to predict for classification methods.
            dmld_hidden: Dimension of hidden vector for DMLD method.
            reduction: Final loss tensor reduction type.
            max_dur: Maximum duration value to cover.
            xe_steps_coef: Ratio of adjusted steps for 'xe-steps' method.
            pad16: True if pad tensors to 16.
        """

        super().__init__()

        self._method = method
        self._num_classes = num_classes
        self._dmld_hidden = dmld_hidden
        self._reduction = reduction
        self._pad16 = pad16

        # Creates XE Steps classes.
        classes = np.arange(num_classes).tolist()
        k, c = 1, num_classes - 1
        while c < max_dur:
            k *= xe_steps_coef
            c += k
            classes.append(int(c))  # noqa
        self._xe_steps_classes = classes
        if self._method == 'xe-steps':
            nemo.logging.info('XE Steps Classes: %s', str(classes))

        # w = torch.arange(num_classes, dtype=torch.float, device=self._device)
        # w = (w + 1).log() + 1
        # w /= w.sum()
        self._weights = None

    def _loss_function(self, dur_true, dur_pred, text_mask):
        if self._method.startswith('l2'):
            if dur_pred.shape[-1] != 1:
                raise ValueError("Wrong `dur_pred` shape.")
            dur_pred = dur_pred.squeeze(-1)

        if self._method == 'l2-log':
            loss = F.mse_loss(dur_pred, (dur_true + 1).float().log(), reduction='none')
        elif self._method == 'l2':
            loss = F.mse_loss(dur_pred, dur_true.float(), reduction='none')
        elif self._method == 'dmld-log':
            # [0, inf] => [0, num_classes - 1]
            dur_true = torch.clamp(dur_true, max=self._num_classes - 1)
            # [0, num_classes - 1] => [0, log(num_classes)]
            dur_true = (dur_true + 1).float().log()
            # [0, log(num_classes)] => [-1, 1]
            dur_true = (torch.clamp(dur_true / math.log(self._num_classes), max=1.0) - 0.5) * 2

            loss = nemo_tts.parts.dmld_loss(dur_pred, dur_true, self._num_classes)
        elif self._method == 'dmld':
            # [0, inf] => [0, num_classes - 1]
            dur_true = torch.clamp(dur_true, max=self._num_classes - 1)
            # [0, num_classes - 1] => [-1, 1]
            dur_true = (dur_true / (self._num_classes - 1) - 0.5) * 2

            loss = nemo_tts.parts.dmld_loss(dur_pred, dur_true, self._num_classes)
        elif self._method == 'xe':
            # [0, inf] => [0, num_classes - 1]
            dur_true = torch.clamp(dur_true, max=self._num_classes - 1)

            loss = F.cross_entropy(
                input=dur_pred.transpose(1, 2), target=dur_true, reduction='none', weight=self._weights,
            )
        elif self._method == 'xe-steps':
            # [0, inf] => [0, xe-steps-num-classes - 1]
            classes = torch.tensor(self._xe_steps_classes, device=dur_pred.device)
            a = dur_true.unsqueeze(-1).repeat(1, 1, *classes.shape)
            b = classes.unsqueeze(0).unsqueeze(0).repeat(*dur_true.shape, 1)
            dur_true = (a - b).abs().argmin(-1)

            loss = F.cross_entropy(input=dur_pred.transpose(1, 2), target=dur_true, reduction='none')
        else:
            raise ValueError("Wrong Method")

        loss *= text_mask.float()
        if self._reduction == 'all':
            loss = loss.sum() / text_mask.sum()
        elif self._reduction == 'batch':
            loss = loss.sum(-1) / text_mask.sum(-1)
            loss = loss.mean()
        else:
            raise ValueError("Wrong Reduction")

        return loss

    @property
    def d_out(self):
        if self._method == 'l2-log':
            return 1
        elif self._method == 'l2':
            return 1
        elif self._method == 'dmld-log':
            return 3 * self._dmld_hidden
        elif self._method == 'dmld':
            return 3 * self._dmld_hidden
        elif self._method == 'xe':
            return self._num_classes
        elif self._method == 'xe-steps':
            # noinspection PyTypeChecker
            return len(self._xe_steps_classes)
        else:
            raise ValueError("Wrong Method")

    def preprocessing(self, tensors):
        if self._method == 'l2-log':
            dur_pred = tensors.dur_pred.squeeze(-1).exp() - 1
        elif self._method == 'l2':
            dur_pred = tensors.dur_pred.squeeze(-1)
        elif self._method == 'dmld-log':
            dur_pred = nemo_tts.parts.dmld_sample(tensors.dur_pred)

            # [-1, 1] => [0, log(num_classes)]
            dur_pred = (dur_pred + 1) / 2 * math.log(self._num_classes)
            # [0, log(num_classes)] => [0, num_classes - 1]
            dur_pred = torch.clamp(dur_pred.exp() - 1, max=self._num_classes - 1)
        elif self._method == 'dmld':
            dur_pred = nemo_tts.parts.dmld_sample(tensors.dur_pred)

            # [-1, 1] => [0, num_classes - 1]
            dur_pred = (dur_pred + 1) / 2 * (self._num_classes - 1)
        elif self._method == 'xe':
            dur_pred = tensors.dur_pred.argmax(-1)
        elif self._method == 'xe-steps':
            dur_pred = tensors.dur_pred.argmax(-1)
            classes = torch.tensor(self._xe_steps_classes, device=dur_pred.device)
            b = classes.unsqueeze(0).unsqueeze(0).repeat(*dur_pred.shape, 1)
            dur_pred = b.gather(-1, dur_pred.unsqueeze(-1)).squeeze(-1)
        else:
            raise ValueError("Wrong Method")

        dur_pred[dur_pred < 0.0] = 0.0
        dur_pred = dur_pred.float().round().long()
        tensors.dur_pred = dur_pred

        return tensors


class TalkNetMelsLoss(LossNM):
    """Neural Module Wrapper for TalkNet mel loss."""

    @property
    @add_port_docs
    def input_ports(self):
        """Returns definitions of module input ports."""
        return dict(
            true=NeuralType(('B', 'D', 'T'), ChannelType()),  # 'BDT' - to fit mels directly.
            pred=NeuralType(('B', 'T', 'D'), ChannelType()),
            mask=NeuralType(('B', 'T'), MaskType()),
            mel_len=NeuralType(('B',), LengthsType(), optional=True),
            dur_true=NeuralType(('B', 'T'), LengthsType(), optional=True),
        )

    @property
    @add_port_docs
    def output_ports(self):
        """Returns definitions of module output ports."""
        return dict(loss=NeuralType(None))

    def __init__(self, reduction='all', pad16=False, doubling=False):
        """Creates instance of TalkNet mels loss calculation.

        Args:
            reduction: Final loss tensor reduction type.
            pad16: True if pad tensors to 16.
            doubling: True if using mel channels doubling trick.
        """

        super().__init__()

        self._reduction = reduction
        self._pad16 = pad16
        self._doubling = doubling

    def _loss_function(self, true, pred, mask, mel_len=None, dur_true=None):
        if mel_len is not None and dur_true is not None:
            if not torch.equal(mel_len, dur_true.sum(-1)) or not torch.equal(mel_len, mask.sum(-1)):
                raise ValueError("Wrong mel length calculation.")

        true = true.transpose(-1, -2)
        if self._pad16:
            true = Ops.pad(true)
        if self._doubling:
            true = Ops.pad(true, to=2)

        loss = F.mse_loss(pred, true, reduction='none').mean(-1)

        if self._pad16:
            mask = Ops.pad(mask)
        if self._doubling:
            mask = Ops.make_mask(mel_len)
            mask = Ops.pad(mask, to=2)
        loss *= mask.float()

        if self._reduction == 'all':
            loss = loss.sum() / mask.sum()
        elif self._reduction == 'batch':
            loss = loss.sum(-1) / mask.sum(-1)
            loss = loss.mean()
        else:
            raise ValueError("Wrong reduction.")

        return loss



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.data.tokenizers.tokenizer_spec import TokenizerSpec

__all__ = ['CharTokenizer']


class CharTokenizer(TokenizerSpec):
    def __init__(self, vocab_path):

        vocab_list = open(vocab_path, "r").readlines()
        self.vocab = {vocab_list[i][0]: i for i in range(len(vocab_list))}
        for special_token in ["<PAD>", "<UNK>", "<BOS>", "<EOS>"]:
            if special_token not in self.vocab:
                self.vocab[special_token] = len(self.vocab)
        self.inv_vocab = {v: k for k, v in self.vocab.items()}
        self.vocab_size = len(self.vocab)
        self.special_tokens = self.tokens_to_ids(["<PAD>", "<UNK>", "<BOS>", "<EOS>"])

    def text_to_tokens(self, text):
        token_candidates = [char for char in text]
        tokens = []
        for token in token_candidates:
            if token in self.vocab:
                tokens.append(token)
            else:
                tokens.append("<UNK>")
        return tokens

    def tokens_to_text(self, tokens):
        return self.ids_to_text(self.tokens_to_ids(tokens))

    def text_to_ids(self, text):
        return [self.vocab[token] for token in self.text_to_tokens(text)]

    def ids_to_text(self, ids):
        ids_ = [id_ for id_ in ids if id_ not in self.special_tokens]
        return "".join(self.ids_to_tokens(ids_))

    def tokens_to_ids(self, tokens):
        return [self.vocab[token] for token in tokens]

    def ids_to_tokens(self, ids):
        return [self.inv_vocab[id] for id in ids]

    @property
    def pad_id(self):
        return self.vocab["<PAD>"]

    @property
    def bos_id(self):
        return self.vocab["<BOS>"]

    @property
    def eos_id(self):
        return self.vocab["<EOS>"]



# =============================================================================
# Copyright 2020 NVIDIA. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

from nemo.collections.nlp.data import GLUEDataset
from nemo.collections.nlp.nm.data_layers.text_datalayer import TextDataLayer
from nemo.core import CategoricalValuesType, ChannelType, NeuralType, RegressionValuesType
from nemo.utils.decorators import add_port_docs

__all__ = ['GlueClassificationDataLayer', 'GlueRegressionDataLayer']


class GlueClassificationDataLayer(TextDataLayer):
    """
    Creates the data layer to use for the GLUE classification tasks,
    more details here: https://gluebenchmark.com/tasks

    All the data processing is done in GLUEDataset.

    Args:
        data_dir (str): data directory path
        tokenizer (TokenizerSpec): text tokenizer.
        max_seq_length (int): maximum allowed length of the text segments .
        processor (DataProcessor): data processor.
        evaluate (bool): true if data layer is used for evaluation. Default: False.
        batch_size (int): batch size in segments
        shuffle (bool): whether to shuffle data or not. Default: False.
        dataset_type (GLUEDataset):
                the dataset that needs to be converted to DataLayerNM
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        input_ids:
            indices of tokens which constitute batches of text segments
        input_type_ids:
            tensor with 0's and 1's to denote the text segment type
        input_mask:
            bool tensor with 0s in place of tokens to be masked
        labels:
            integer indices for sentence classication prediction
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
            "labels": NeuralType(tuple('B'), CategoricalValuesType()),
        }

    def __init__(
        self,
        data_dir,
        tokenizer,
        max_seq_length,
        processor,
        evaluate=False,
        shuffle=False,
        batch_size=64,
        dataset_type=GLUEDataset,
        use_data_cache=False,
    ):
        dataset_params = {
            'data_dir': data_dir,
            'output_mode': 'classification',
            'processor': processor,
            'evaluate': evaluate,
            'tokenizer': tokenizer,
            'max_seq_length': max_seq_length,
            'use_data_cache': use_data_cache,
        }
        super().__init__(dataset_type, dataset_params, batch_size, shuffle)


class GlueRegressionDataLayer(TextDataLayer):
    """
    Creates the data layer to use for the GLUE STS-B regression task,
    more details here: https://gluebenchmark.com/tasks

    All the data processing is done in GLUEDataset.

    Args:
        data_dir (str): data directory path
        tokenizer (TokenizerSpec): text tokenizer.
        max_seq_length (int): maximum allowed length of the text segments .
        processor (DataProcessor): data processor.
        evaluate (bool): true if data layer is used for evaluation. Default: False.
        batch_size (int): batch size in segments
        shuffle (bool): whether to shuffle data or not. Default: False.
        dataset_type (GLUEDataset):
                the dataset that needs to be converted to DataLayerNM
    """

    @property
    @add_port_docs()
    def output_ports(self):
        """Returns definitions of module output ports.

        input_ids:
            indices of tokens which constitute batches of text segments
        input_type_ids:
            tensor with 0's and 1's to denote the text segment type
        input_mask:
            bool tensor with 0s in place of tokens to be masked
        labels:
            float for sentence regression prediction
        """
        return {
            "input_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_type_ids": NeuralType(('B', 'T'), ChannelType()),
            "input_mask": NeuralType(('B', 'T'), ChannelType()),
            "labels": NeuralType(tuple('B'), RegressionValuesType()),
        }

    def __init__(
        self,
        data_dir,
        tokenizer,
        max_seq_length,
        processor,
        evaluate=False,
        shuffle=False,
        batch_size=64,
        dataset_type=GLUEDataset,
        use_data_cache=False,
    ):
        dataset_params = {
            'data_dir': data_dir,
            'output_mode': 'regression',
            'processor': processor,
            'evaluate': evaluate,
            'tokenizer': tokenizer,
            'max_seq_length': max_seq_length,
            'use_data_cache': use_data_cache,
        }

        super().__init__(dataset_type, dataset_params, batch_size, shuffle)



# Copyright (c) 2019-, NVIDIA CORPORATION. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['WeightShareTransform', 'NeuralModule', 'PretrainedModelInfo', 'ModuleType', 'OperationMode']

import uuid
from abc import abstractmethod
from collections import namedtuple
from enum import Enum
from inspect import getargvalues, getfullargspec, stack
from os import path
from typing import Any, Dict, List, Optional, Set, Tuple

from ruamel.yaml import YAML

from nemo.core.neural_factory import NeuralModuleFactory, OperationMode
from nemo.core.neural_interface import NeuralInterface
from nemo.core.neural_types import NeuralPortNameMismatchError, NeuralType, NmTensor
from nemo.package_info import __version__ as nemo_version
from nemo.utils import logging
from nemo.utils.decorators.deprecated import deprecated
from nemo.utils.neural_graph.connection import StepModulePort

YAML = YAML(typ='safe')


class ModuleType(Enum):
    """ Back-end independent module types """

    module = 0
    datalayer = 1
    trainable = 2
    loss = 3
    nontrainable = 4


class WeightShareTransform(Enum):
    """When sharing parameters, what kind of transform to apply."""

    SAME = 0
    TRANSPOSE = 1


PretrainedModelInfo = namedtuple(
    "PretrainedModleInfo", ("pretrained_model_name", "description", "parameters", "location"),
)


class NeuralModule(NeuralInterface):
    """
    Abstract class that every Neural Module must inherit from.
    """

    def __init__(self, name=None):
        # Initialize the inferface.
        super().__init__()

        # Retrieve dictionary of parameters (keys, values) passed to init.
        self._init_params = self.__extract_init_params()

        # Get object UUID.
        self._uuid = str(uuid.uuid4())

        # Register module and store the generated name.
        self._name = self._app_state.register_module(self, name)

        # Set "module" type as default.
        self._type = ModuleType.module

        # Set "both" as default operation mode.
        self._operation_mode = OperationMode.both

        # Get default factory.
        self._factory = NeuralModuleFactory.get_default_factory()
        if self._factory is None:
            raise NameError(
                "A Neural Module was instantiated without first creating a Neural Factory. Please instantiate the "
                "NeuralModuleFactory class first."
            )

        # Set module properties from factory else use defaults
        self._placement = self._factory.placement
        # If one needs to change that should override it manually.

        # Optimization level.
        self._opt_level = self._factory.optim_level

    @property
    def init_params(self) -> Dict[str, Any]:
        """
        Property returning parameters used to instantiate the module.

        Returns:
            Dictionary containing parameters used to instantiate the module.
        """
        return self._init_params

    def __extract_init_params(self) -> Dict[str, Any]:
        """
        Retrieves the dictionary of of parameters (keys, values) passed to constructor of a class derived
        (also indirectly) from the Neural Module class.

        Returns:
            Dictionary containing parameters passed to init().
        """
        # Get names of arguments of the original module init method.
        to_set_params = getfullargspec(type(self).__init__).args
        to_set_params.remove("self")

        # Create empty list of init params.
        init_params = {}

        # Get the frame "call context".
        for frame in stack()[1:]:
            # Get the current call arguments.
            localvars = getargvalues(frame[0])

            # Fill the parameters with call arguments.
            for key in to_set_params:
                if key in localvars.args:
                    init_params[key] = localvars.locals[key]

            # Remove all set keys.
            for key in init_params:
                if key in to_set_params:
                    to_set_params.remove(key)

            # Check if we have set everything.
            if len(to_set_params) == 0:
                break

        # Make sure that we collected ALL (and ONLY) the signature params - if not, then there is a BUG!
        if len(to_set_params) != 0:
            raise ValueError(
                "Could not collect all the signature params! "
                f"Please file a bug on GitHub with the current stack trace so that it can be reproduced."
            )

        # print("! init_params of {}: {}\n".format(type(self).__name__, init_params))

        # Return parameters.
        return init_params

    def __validate_params(self, params: Dict[str, Any]) -> bool:
        """
        Checks whether dictionary contains parameters being primitive types (string, int, float etc.)
        or (lists of)+ primitive types.

        Args:
            params: dictionary of parameters.
        Returns:
            True if all parameters were ok, False otherwise.
        """
        ok = True

        # Iterate over parameters and check them one by one.
        for key, variable in params.items():
            if not self.__is_of_allowed_type(variable):
                logging.warning(
                    "Parameter '{}' contains a variable '{}' of type '{}' which is not allowed.".format(
                        key, variable, type(variable)
                    )
                )
                ok = False

        # Return the result.
        return ok

    def __is_of_allowed_type(self, var) -> bool:
        """
        A recursive function that checks if a given variable is of allowed type.

        Args:
            pretrained_model_name (str): name of pretrained model to use in order.
        Returns:
            True if all parameters were ok, False otherwise.
        """
        # Special case: None is also allowed.
        if var is None:
            return True

        var_type = type(var)

        # If this is list - check its elements.
        if var_type == list:
            for list_var in var:
                if not self.__is_of_allowed_type(list_var):
                    return False

        # If this is dict - check its elements.
        elif var_type == dict:
            for _, dict_var in var.items():
                if not self.__is_of_allowed_type(dict_var):
                    return False

        elif var_type not in (str, int, float, bool):
            return False

        # Well, seems that everything is ok.
        return True

    def export_to_config(self, config_file: str):
        """
        A function that exports module "configuration" (i.e. init parameters) to a YAML file.

        Args:
            config_file: path (absolute or relative) and name of the config file (YML)
        Raises:
            ValueError: An error occurred and  parameters coudn't be exported.
        """
        # Greate an absolute path.
        abs_path_file = path.expanduser(config_file)

        # Serialize the module.
        to_export = self.serialize()

        # All parameters are ok, let's export.
        with open(abs_path_file, 'w') as outfile:
            YAML.dump(to_export, outfile)

        logging.info(
            "Configuration of module `{}` ({}) exported to '{}'".format(self.name, type(self).__name__, abs_path_file)
        )

    def serialize(self) -> Dict[str, Any]:
        """
        A method serializing the whole Neural module (into a dictionary).

        Returns:
            Dictionary containing a "serialized" module.
        """
        # Create a dictionary representing the serialized object.
        serialized_module = {}

        # Add "header" with module "specification".
        serialized_module["header"] = self.__serialize_header()

        # Add init parameters.
        serialized_module["init_params"] = self._serialize_configuration()

        # Return the dictionary.
        return serialized_module

    def __serialize_header(self) -> Dict[str, Any]:
        """
        A protected method that creates a header stored later in the configuration file.

        Returns:
            Dictionary containing a header with module specification.
        """

        # Get module "full specification".
        module_full_spec = str(self.__module__) + "." + str(self.__class__.__qualname__)
        module_class_name = type(self).__name__
        # print(module_full_spec)

        # Check whether module belongs to a collection.
        spec_list = module_full_spec.split(".")

        # Do not check Neural Modules from unit tests.
        if spec_list[0] == "tests":
            # Set collection variables.
            collection_type = "tests"
            collection_version = None
        else:
            # Check if component belongs to any collection
            if len(spec_list) < 3 or (spec_list[0] != "nemo" and spec_list[1] != "collection"):
                logging.warning(
                    "Module `{}` does not belong to any collection. This won't be allowed in the next release.".format(
                        module_class_name
                    )
                )
                collection_type = "unknown"
                collection_version = None
            else:
                # Ok, set collection.
                collection_type = spec_list[2]
                collection_version = None
                # TODO: to be SET!
                # print(getattr("nemo.collections.nlp", __version__))

        # Create a "header" with module "specification".
        header = {
            "nemo_core_version": nemo_version,
            "collection_type": collection_type,
            "collection_version": collection_version,
            # "class": module_class_name, # Operating only on full_spec now.
            "full_spec": module_full_spec,
        }
        return header

    def _serialize_configuration(self) -> Dict[str, Any]:
        """
        A function that serializes the module "configuration (i.e. init parameters) to a dictionary.

        ..note:
            Thus functions should be overloaded when writing a custom module import/export.

        Returns:
            A "serialized" dictionary with module configuration.
        Raises:
            A ValueError exception in case then parameters coudn't be exported.
        """
        # Check if generic export will work.
        if not self.__validate_params(self._init_params):
            raise ValueError(
                "Generic configuration export enables to use of parameters of primitive types (string, int, float) "
                F"or (lists of/dicts of) primitive types. Please implement your own custom `export_to_config()` and "
                F"`import_from_config()` methods for your custom Module class."
            )
        # In this case configuration = init parameters.
        return self._init_params

    @classmethod
    def import_from_config(
        cls, config_file: str, section_name: str = None, name: str = None, overwrite_params: Dict = {}
    ) -> 'NeuralModule':
        """
        Class method importing the configuration file.
        Raises an ImportError exception when config file is invalid or
        incompatible (when called from a particular class).

        Args:
            config_file: path (absolute or relative) and name of the config file (YML)
            section_name: section in the configuration file storing module configuration (optional, DEFAULT: None)
            name: name of the module that will overwrite the name in the `init_params` (optional, DEFAULT: None)
            overwrite_params: Dictionary containing parameters that will be added to or overwrite (!)
            the default init parameters loaded from the configuration file (the module "init_params" section).
        Returns:
            Instance of the created NeuralModule object.
        """
        logging.info("Loading configuration of a new Neural Module from the `{}` file".format(config_file))

        # Validate the content of the configuration file (its header).
        loaded_config = cls.__validate_config_file(config_file, section_name)

        # "Deserialize" the module.
        obj = cls.deserialize(loaded_config, name, overwrite_params)

        # Return the new module.
        return obj

    @classmethod
    def __validate_config_file(cls, config_file: str, section_name: str = None) -> Dict[str, Any]:
        """
        Class method validating whether the config file has a proper content (sections, specification etc.).
        Raises an ImportError exception when config file is invalid or
        incompatible (when called from a particular class).

        Args:
            config_file: path (absolute or relative) and name of the config file (YML)
            section_name: section in the configuration file storing module configuration (optional, DEFAULT: None)
        Returns:
            A loaded configuration file (dictionary).
        """
        # Greate an absolute path.
        abs_path_file = path.expanduser(config_file)

        # Open the config file.
        with open(abs_path_file, 'r') as stream:
            loaded_config = YAML.load(stream)

        # Check section.
        if section_name is not None:
            if section_name not in loaded_config:
                raise ImportError(
                    "The loaded config `{}` doesn't contain the indicated `{}` section".format(
                        config_file, section_name
                    )
                )
            # Section exists - use only it for configuration.
            loaded_config = loaded_config[section_name]

        # Make sure that the config is valid.
        if "header" not in loaded_config:
            raise ImportError("The loaded config `{}` doesn't contain the `header` section".format(config_file))

        if "init_params" not in loaded_config:
            raise ImportError("The loaded config `{}` doesn't contain the `init_params` section".format(config_file))

        # Parse the "full specification".
        spec_list = loaded_config["header"]["full_spec"].split(".")

        # Check if config contains data of a compatible class.
        if not issubclass(cls.__deserialize_header(loaded_config["header"]), cls):
            txt = "The loaded file `{}` contains configuration of ".format(config_file)
            txt = txt + "`{}` thus cannot be used for instantiation of an object of type `{}`".format(
                spec_list[-1], cls.__name__
            )
            raise ImportError(txt)

        # Success - return configuration.
        return loaded_config

    @classmethod
    def deserialize(
        cls, configuration: Dict[str, Any], name: str = None, overwrite_params: Dict[str, Any] = {}
    ) -> 'NeuralModule':
        """
        Class method instantiating the neural module object based on the configuration (dictionary).

        Args:
            configuration: Dictionary containing proper "header" and "init_params" sections.

            name: name of the module that will overwrite the name in the `init_params` (optional, DEFAULT: None)

            overwrite_params: Dictionary containing parameters that will be added to or overwrite (!)
            the default init parameters loaded from the configuration file (the module "init_params" section).

        Returns:
            Instance of the created NeuralModule object.
        """
        # Deserialize header - get object class.
        module_class = cls.__deserialize_header(configuration["header"])

        # Update parameters with additional ones.
        configuration["init_params"].update(overwrite_params)

        # Override module name in init_params using the logic:
        #  * section_name if not none overrides init_params.name first (skipped for now, TOTHINK!)
        #  * name (if None) overrides init_params.name
        if name is not None:
            configuration["init_params"]["name"] = name

        # Get init parameters.
        init_params = cls._deserialize_configuration(configuration["init_params"])

        # Create the module instance.
        new_module = module_class(**init_params)
        logging.info(
            "Instantiated a new Neural Module named `{}` of type `{}`".format(
                new_module.name, type(new_module).__name__
            )
        )

        # Return the module instance.
        return new_module

    @classmethod
    def __deserialize_header(cls, serialized_header: Dict[str, Any]):
        """
        Method deserializes the header and extracts the module class.

        Args:
            serialized_header: Dictionary containing module header.
        Returns:
            Class of the module to be created.
        """
        # Parse the "full specification".
        spec_list = serialized_header["full_spec"].split(".")

        # Get module class from the "full specification".
        mod_obj = __import__(spec_list[0])
        for spec in spec_list[1:]:
            mod_obj = getattr(mod_obj, spec)

        # Return "class".
        return mod_obj

    @classmethod
    def _deserialize_configuration(cls, serialized_init_params: Dict[str, Any]):
        """
        A function that deserializes the module "configuration (i.e. init parameters).

        ..note:
            Thus functions should be overloaded when writing a custom module import/export.

        Args:
            serialized_init_params: List of init parameters loaded from the file.
        Returns:
            A "deserialized" list with init parameters.
        """
        # In this case configuration = init parameters.
        return serialized_init_params

    @property
    @abstractmethod
    def input_ports(self) -> Dict[str, NeuralType]:
        """
        Returns definitions of module input ports

        Returns:
          A dictionary containing module's input ports (names, NeuralTypes) mapping.
        """

    @property
    @abstractmethod
    def output_ports(self) -> Dict[str, NeuralType]:
        """
        Returns definitions of module output ports

        Returns:
          A dictionary containing module's output ports (names, NeuralTypes) mapping.
        """

    @property
    def _disabled_deployment_input_ports(self) -> Set[str]:
        """Returns names of input ports that will not be included in an export

        Returns:
          A (set) of module's input port names that are not exportable
        """
        return set()

    @property
    def _disabled_deployment_output_ports(self) -> Set[str]:
        """Returns names of output ports that will not be included in an export

        Returns:
          A (set) of module's output port names that are not exportable
        """
        return set()

    def _prepare_for_deployment(self) -> None:
        """Patch the module if required to prepare for deployment

        Returns:
            (Optional) input and output example tensors
        """
        return None, None

    @property
    def operation_mode(self):
        """ Returns the operation mode. """
        return self._operation_mode

    @property
    def type(self):
        """ Returns the type of module. """
        return self._type

    @operation_mode.setter
    def operation_mode(self, operation_mode: OperationMode):
        """ Sets the operation mode. """
        self._operation_mode = operation_mode

    @staticmethod
    def pretrained_storage():
        return ''

    def __call__(self, **kwargs):
        """This method allows objects to be called with their port names

        Args:
          kwargs: Input ports and their values. For example:
          ...
          mymodule1 = Subclass1_of_NeuralModule(...)
          mymodule2 = Subclass2_of_NeuralModule(...)
          ...
          out_port1, out_port2 = mymodule1(input_port1=value1,
          input_port2=value2,
          input_port3=value3)
          out_port11 = mymodule2(input_port1=out_port2)
          ...

        Returns:
          NmTensor object or tuple of NmTensor objects
        """
        # print(" Neural Module:__call__")

        # Set the operation mode of the outer graph.
        self.operation_mode = self._app_state.active_graph.operation_mode
        # The input and output ports definitions can potentially depend on the operation mode!

        # Record the operation (i.e. add a single module).
        step_number = self._app_state.active_graph.record_step(self)

        ###### PROCESS INPUTS. ######
        # Iterate through all passed parameters.
        for port_name, port_content in kwargs.items():
            # Make sure that passed arguments corresponds to one of the input port names.
            if port_name not in self.input_ports.keys():
                raise NeuralPortNameMismatchError(port_name)

            # At that point the input can be one of three types:
            # * NeuralGraph -> bind port using the default name and type.
            # * GraphInput -> check definition, if ok bind port.
            # * NmTensor -> check definition, add self as a "consumer" of a tensor (produced by other module).

            # Check what was actually passed.
            if type(port_content).__name__ == "NeuralGraph":
                # Make sure that port_content is the currently active graph!
                if port_content is not self._app_state.active_graph:
                    raise ConnectionError("Ports can be bound only by passing the active graph object!")
                # Create an alias so the logic will be more clear.
                active_graph = port_content

                # This case: we are nesting one graph into another and must bind input port of one graph in another!
                # So generally we must "copy" the of thus module to graog (the inverted logic!).

                # Copy the port "definition" (i.e. is NeuralType) using the same port name.
                active_graph.inputs[port_name] = self.input_ports[port_name]

                # Bind the neural graph input port, i.e. remember that a given graph port should pass data
                # to THIS module-port (when it finally will be connected).
                active_graph.inputs[port_name].bind(StepModulePort(step_number, self.name, port_name))

                # Please note that there are no "consumers" here - this is a "pure binding".

            elif type(port_content).__name__ == "GraphInput":

                # Check if GraphInput belongs to the active graph !
                own_port = False
                for gcontent in self._app_state.active_graph.inputs.values():
                    if gcontent is port_content:
                        own_port = True
                        break
                if not own_port:
                    raise NeuralPortNameMismatchError(port_name)

                # Compare input port definition with the received definition.
                self.input_ports[port_name].compare_and_raise_error(
                    self.__class__.__name__, port_name, port_content.ntype
                )

                # Bind the neural graph input port, i.e. remember that a given graph port should pass data
                # to THIS module-port (when it finally will be connected).
                port_content.bind(StepModulePort(step_number, self.name, port_name))

                # Please note that there are no "consumers" here - this is a "pure binding".

            elif type(port_content) is NmTensor:
                # Compare input port definition with the received definition.
                self.input_ports[port_name].compare_and_raise_error(self.__class__.__name__, port_name, port_content)

                # Ok, the goal here is to actually "connect": add self (module) as "consumer" to the input tensor.
                port_content.add_consumer(StepModulePort(step_number, self.name, port_name))
            else:
                raise TypeError(
                    "Input '{}' must be of one of three types: NeuralGraph, GraphInput or NmTensor".format(port_name)
                )

        ###### PRODUCE OUTPUTS. ######
        output_port_defs = self.output_ports
        # Create output tensors.
        if len(output_port_defs) == 1:
            # Get port name and type.
            out_name = list(output_port_defs)[0]
            out_type = output_port_defs[out_name]

            # Create a single returned tensor.
            results = NmTensor(producer=self, producer_args=kwargs, output_port_name=out_name, ntype=out_type,)

            # Bind the "default" output ports.
            self._app_state.active_graph.bind_outputs(results)
        else:
            # Create output tensors.
            output_tensors = []
            for out_name, out_type in output_port_defs.items():
                output_tensors.append(
                    NmTensor(producer=self, producer_args=kwargs, output_port_name=out_name, ntype=out_type,)
                )

            # Create a named tuple type enabling to access outputs by attributes (e.g. out.x).
            output_class_name = f'{self.__class__.__name__}Output'
            result_type = namedtuple(typename=output_class_name, field_names=output_port_defs.keys())

            # Create the returned tuple object.
            results = result_type(*output_tensors)

            # Bind the output tensors.
            self._app_state.active_graph.bind_outputs(output_tensors)

        # Return the results.
        return results

    def __str__(self):
        return self.__class__.__name__

    @abstractmethod
    def get_weights(self) -> Optional[Dict[(str, bool)]]:
        """Returns NeuralModule's weights copy.

        Returns:
          Dictionary of name -> (weights, trainable)"""
        pass

    @abstractmethod
    def set_weights(
        self,
        name2weight: Dict[(str, Tuple[str, bool])],
        name2name_and_transform: Dict[(str, Tuple[str, WeightShareTransform])] = None,
    ):
        """Sets weight from given values. For every named weight in
        name2weight,
        if weight with the same name is found in the model, it will be set to
        found value.

        WARNING: This will NOT tie weights. It will copy values.

        If ``name2name_and_transform`` is provided then if will set weights
        using
        name mapping and transform. For example, suppose ``objec1.X = 3x5
        weight``.
        Then, if ``name2name_and_transform['X']=('Y',
        WeightShareTransform.TRANSPOSE)``
        and ``Y`` is 5x3 weight and ``name2weight['Y']=Y. Then:
        ``object1.set_weights(name2weight, name2name_and_transform)`` will
        set object1.X=transpose(Y).

        Args:
          name2weight (dict): dictionary of name to (weight, trainable).
          Typically this is output of get_weights method.
          name2name_and_transform: mapping from name -> (name, transform)
        """
        pass

    @staticmethod
    def list_pretrained_models() -> Optional[List[PretrainedModelInfo]]:
        """List all available pre-trained models (e.g. weights) for this NM.

        Returns:
            A list of PretrainedModelInfo tuples.
            The pretrained_model_name field of the tuple can be used to
            retrieve pre-trained model's weights (pass it as
            pretrained_model_name argument to the module's constructor)
        """
        return None

    def get_config_dict_and_checkpoint(self, pretrained_model_name):
        """WARNING: This part is work in progress"""
        return None

    @abstractmethod
    def tie_weights_with(
        self,
        module,
        weight_names=List[str],
        name2name_and_transform: Dict[(str, Tuple[str, WeightShareTransform])] = None,
    ):
        """Ties weights between self and module. For every weight name in
        weight_names, if weight with the same name is found in self, it will
        be tied
        with a same weight from ``module``.

        WARNING: Once weights are tied, updates to one weights's weights
        will affect
        other module's weights.


        If ``name2name_and_transform`` is provided then if will set weights
        using
        name mapping and transform. For example, suppose ``objec1.X = 3x5
        weights``
        and ``object2.Y = 5x3 weights``. Then these weights can be tied like
        this:

        .. code-block:: python

          object1.tie_weights_with(object2, weight_names=['X'],
          name2name_and_transform =
          { 'X': ('Y', WeightShareTransform.TRANSPOSE)})


        Args:
            module: with which module to tie weights
            weight_names (List[str]): list of self weights' names
            name2name_and_transform: mapping from name -> (name, transform)
        """
        pass

    def is_trainable(self) -> bool:
        """
        Checks if NeuralModule is trainable.
        A NeuralModule is trainable IFF it contains at least one trainable
        weight

        Returns:
          True if module has trainable weights, False otherwise
        """
        weights = self.get_weights()
        if weights is None:
            return False
        for name, w in weights.items():
            if w[1]:
                return True
        return False

    @abstractmethod
    def save_to(self, path: str):
        """Save module state to file.

        Args:
          path (string): path to while where to save.
        """
        pass

    @abstractmethod
    def restore_from(self, path: str):
        """Restore module's state from file.

        Args:
          path (string): path to where to restore from.
        """
        pass

    @abstractmethod
    def freeze(self, weights: Set[str] = None):
        """Freeze weights

        Args:
          weights (set): set of weight names to freeze
          If None, all weights are freezed.
        """
        pass

    @abstractmethod
    def unfreeze(self, weights: Set[str] = None):
        """Unfreeze weights

        Args:
          weights (set): set of weight names to unfreeze
          If None, all weights are unfreezed.
        """
        pass

    @property
    def placement(self):
        """Module's placement. Currently CPU or GPU.
        DataParallel and ModelParallel will come later.

        Returns:
          (DeviceType) Device where NM's weights are located
        """
        return self._placement

    @property
    @deprecated(version=0.11)
    def local_parameters(self) -> Optional[Dict]:
        """Get module's parameters

        Returns:
          module's parameters
        """
        return self._init_params
        # return self._local_parameters

    @property
    def unique_instance_id(self):
        """A unique instance id for this object

        Returns:
          A uniq uuid which can be used to identify this object
        """
        return self._uuid

    @property
    def factory(self):
        """ Neural module factory which created this module
        Returns: NeuralModuleFactory instance or None
        """
        return self._factory

    @property
    @abstractmethod
    def num_weights(self):
        """Number of module's weights
        """
        pass



# ! /usr/bin/python
# -*- coding: utf-8 -*-
# =============================================================================
# Copyright (c) 2020, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# =============================================================================

import pytest

from nemo.backends.pytorch.tutorials import MSELoss, RealFunctionDataLayer, TaylorNet
from nemo.core.neural_types import NeuralTypeComparisonResult


@pytest.mark.usefixtures("neural_factory")
class TestNmTensor:
    @pytest.mark.unit
    def test_nm_tensors_producer_args(self):
        """
            Tests whether nmTensors are correct - producers and their args.
        """
        # Create modules.
        data_source = RealFunctionDataLayer(n=100, batch_size=1)
        trainable_module = TaylorNet(dim=4)
        loss = MSELoss()

        # Create the graph by connnecting the modules.
        x, y = data_source()
        y_pred = trainable_module(x=x)
        loss_tensor = loss(predictions=y_pred, target=y)

        # check producers' bookkeeping
        assert loss_tensor.producer_name == loss.name
        assert loss_tensor.producer_args == {"predictions": y_pred, "target": y}
        assert y_pred.producer_name == trainable_module.name
        assert y_pred.producer_args == {"x": x}
        assert y.producer_name == data_source.name
        assert y.producer_args == {}
        assert x.producer_name == data_source.name
        assert x.producer_args == {}

    @pytest.mark.unit
    def test_simple_train_named_output(self):
        """ Test named output """
        data_source = RealFunctionDataLayer(n=10, batch_size=1)
        # Get data
        data = data_source()

        # Check output class naming coherence.
        assert type(data).__name__ == 'RealFunctionDataLayerOutput'

        # Check types.
        assert data.x.compare(data_source.output_ports["x"]) == NeuralTypeComparisonResult.SAME
        assert data.y.compare(data_source.output_ports["y"]) == NeuralTypeComparisonResult.SAME

    @pytest.mark.unit
    def test_nm_tensors_producer_consumers(self):
        """
            Tests whether nmTensors are correct - checking producers and consumers.
        """
        # Create modules.
        data_source = RealFunctionDataLayer(n=10, batch_size=1, name="source")
        trainable_module = TaylorNet(dim=4, name="tm")
        loss = MSELoss(name="loss")
        loss2 = MSELoss(name="loss2")

        # Create the graph by connnecting the modules.
        x, y = data_source()
        y_pred = trainable_module(x=x)
        lss = loss(predictions=y_pred, target=y)
        lss2 = loss2(predictions=y_pred, target=y)

        # Check tensor x producer and consumers.
        p = x.producer_step_module_port
        cs = x.consumers
        assert p.module_name == "source"
        assert p.port_name == "x"
        assert len(cs) == 1
        assert cs[0].module_name == "tm"
        assert cs[0].port_name == "x"

        # Check tensor y producer and consumers.
        p = y.producer_step_module_port
        cs = y.consumers
        assert p.module_name == "source"
        assert p.port_name == "y"
        assert len(cs) == 2
        assert cs[0].module_name == "loss"
        assert cs[0].port_name == "target"
        assert cs[1].module_name == "loss2"
        assert cs[1].port_name == "target"

        # Check tensor y_pred producer and consumers.
        p = y_pred.producer_step_module_port
        cs = y_pred.consumers
        assert p.module_name == "tm"
        assert p.port_name == "y_pred"
        assert len(cs) == 2
        assert cs[0].module_name == "loss"
        assert cs[0].port_name == "predictions"
        assert cs[1].module_name == "loss2"
        assert cs[1].port_name == "predictions"

    @pytest.mark.unit
    def test_nm_tensors_types(self):
        """
            Tests whether nmTensors are correct - checking type property.
        """
        # Create modules.
        data_source = RealFunctionDataLayer(n=10, batch_size=1)
        trainable_module = TaylorNet(dim=4)
        loss = MSELoss()

        # Create the graph by connnecting the modules.
        x, y = data_source()
        y_pred = trainable_module(x=x)
        lss = loss(predictions=y_pred, target=y)

        # Check types.
        assert x.ntype.compare(data_source.output_ports["x"]) == NeuralTypeComparisonResult.SAME
        assert y.ntype.compare(data_source.output_ports["y"]) == NeuralTypeComparisonResult.SAME
        assert y_pred.ntype.compare(trainable_module.output_ports["y_pred"]) == NeuralTypeComparisonResult.SAME
        assert lss.ntype.compare(loss.output_ports["loss"]) == NeuralTypeComparisonResult.SAME
